
from pyfmu.fmi2 import get_default_initial, validate_vc,Fmi2Causality, Fmi2Initial, Fmi2Variability

from pytest import raises

# check fmi2 spec page 48-49 for expected behavior

A = {'default' : Fmi2Initial.exact, 'possible': {Fmi2Initial.exact}}
B = {'default' : Fmi2Initial.calculated, 'possible': {Fmi2Initial.approx, Fmi2Initial.calculated}}
C = {'default' : Fmi2Initial.calculated, 'possible': {Fmi2Initial.exact, Fmi2Initial.approx, Fmi2Initial.calculated}}
D = {'default' : None, 'possible': {None}}
E = {'default' : None, 'possible': {None}} 


combinations = [
    {'v': Fmi2Variability.constant, 'c': Fmi2Causality.parameter, 'valid' : False},
    {'v': Fmi2Variability.constant, 'c': Fmi2Causality.calculatedParameter, 'valid' : False},
    {'v': Fmi2Variability.constant, 'c': Fmi2Causality.input, 'valid' : False},
    {'v': Fmi2Variability.constant, 'c': Fmi2Causality.output, 'valid' : True, 'initial' : A},
    {'v': Fmi2Variability.constant, 'c': Fmi2Causality.local, 'valid' : True, 'initial' : A},
    {'v': Fmi2Variability.constant, 'c': Fmi2Causality.independent, 'valid' : False},

    {'v': Fmi2Variability.fixed, 'c': Fmi2Causality.parameter, 'valid' : True, 'initial' : A},
    {'v': Fmi2Variability.fixed, 'c': Fmi2Causality.calculatedParameter, 'valid' : True, 'initial' : B},
    {'v': Fmi2Variability.fixed, 'c': Fmi2Causality.input, 'valid' : False},
    {'v': Fmi2Variability.fixed, 'c': Fmi2Causality.output, 'valid' : False},
    {'v': Fmi2Variability.fixed, 'c': Fmi2Causality.local, 'valid' : True, 'initial' : B},
    {'v': Fmi2Variability.fixed, 'c': Fmi2Causality.independent, 'valid' : False},

    {'v': Fmi2Variability.tunable, 'c': Fmi2Causality.parameter, 'valid' : True, 'initial' : A},
    {'v': Fmi2Variability.tunable, 'c': Fmi2Causality.calculatedParameter, 'valid' : True, 'initial' : B},
    {'v': Fmi2Variability.tunable, 'c': Fmi2Causality.input, 'valid' : False},
    {'v': Fmi2Variability.tunable, 'c': Fmi2Causality.output, 'valid' : False},
    {'v': Fmi2Variability.tunable, 'c': Fmi2Causality.local, 'valid' : True, 'initial' : B},
    {'v': Fmi2Variability.tunable, 'c': Fmi2Causality.independent, 'valid' : False},

    {'v': Fmi2Variability.discrete, 'c': Fmi2Causality.parameter, 'valid' : False},
    {'v': Fmi2Variability.discrete, 'c': Fmi2Causality.calculatedParameter, 'valid' : False},
    {'v': Fmi2Variability.discrete, 'c': Fmi2Causality.input, 'valid' : True, 'initial' : D},
    {'v': Fmi2Variability.discrete, 'c': Fmi2Causality.output, 'valid' : True, 'initial' : C},
    {'v': Fmi2Variability.discrete, 'c': Fmi2Causality.local, 'valid' : True, 'initial' : C},
    {'v': Fmi2Variability.discrete, 'c': Fmi2Causality.independent, 'valid' : False},

    {'v': Fmi2Variability.continuous, 'c': Fmi2Causality.parameter, 'valid' : False},
    {'v': Fmi2Variability.continuous, 'c': Fmi2Causality.calculatedParameter, 'valid' : False},
    {'v': Fmi2Variability.continuous, 'c': Fmi2Causality.input, 'valid' : True, 'initial' : D},
    {'v': Fmi2Variability.continuous, 'c': Fmi2Causality.output, 'valid' : True, 'initial' : C},
    {'v': Fmi2Variability.continuous, 'c': Fmi2Causality.local, 'valid' : True, 'initial' : C},
    {'v': Fmi2Variability.continuous, 'c': Fmi2Causality.independent, 'valid' : True, 'initial' : E}
]

def test_validate_vc():
    
    for c in combinations:
        va = c['v']
        ca = c['c']
        is_valid_expect = c['valid']
        ret = validate_vc(va,ca)

        if(is_valid_expect):
            assert(ret is None)
        else:
            assert(ret != None)

            


def test_get_default_initial():
    
    for c in combinations:
        
        va = c['v']
        ca = c['c']
        is_valid = c['valid']

        if(is_valid):
            default_actual = get_default_initial(va,ca)
            default_expect = c['initial']['default']
            
            assert(default_expect == default_actual)



from zipfile import ZipFile, ZIP_DEFLATED
import os
from os.path import dirname, isdir, isfile, join, normpath
from pathlib import Path
import subprocess

def zipdir(inDir: str, outDir: str):

    if(not isdir(inDir)):
        raise ValueError("Input path does not correspond to a directory!")

    with ZipFile(outDir, 'w', ZIP_DEFLATED) as zf:
        for root, dirs, files in os.walk(inDir):
            for file in files:
                p = os.path.relpath(os.path.join(root, file),
                                    os.path.join(inDir, '..'))
                zf.write(p)


class cd:
    """Context manager for changing the current working directory"""

    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)

def has_java():
    """Checks if java is available in the system's path.
    
    Returns:
        [bool] -- true if java is available, otherwise false.
    """
    try:
        subprocess.run(['java','-v'])
    except Exception:
        return False
    
    return True

def has_fmpy() -> bool:
    """Checks if FMPy is available
    
    Returns:
        bool -- true if FMPy is available, otherwise false.
    """
    try:
        import fmpy
    except:
        return False

    return True

if __name__ == "__main__":

    from os.path import join, dirname

    inPath = join(dirname(__file__), "test")
    outpath = join(dirname(__file__), "out.zip")

    zipdir(inPath, outpath)



from os.path import join
from xml.dom import minidom
from xml.etree.ElementTree import ElementTree
import datetime
import io
import uuid
import xml.etree.ElementTree as ET

from jinja2 import Template

from pyfmu.fmi2 import Fmi2Causality


def extract_model_description_v2(fmu_instance) -> str:
    
    data_time_obj = datetime.datetime.now()
    date_str_xsd = datetime.datetime.strftime(data_time_obj, '%Y-%m-%dT%H:%M:%SZ')

    uid = str(uuid.uuid4())

    fmd = ET.Element("fmiModelDescription")
    fmd.set("fmiVersion","2.0")
    fmd.set("modelName",fmu_instance.modelName)
    fmd.set("guid",uid)
    fmd.set('author',fmu_instance.author)
    fmd.set('generationDateAndTime', date_str_xsd)
    fmd.set('variableNamingConvention', 'structured')
    fmd.set("generationTool", 'pyfmu')

    cs = ET.SubElement(fmd,'CoSimulation')
    cs.set("modelIdentifier", 'pyfmu')
    cs.set('needsExecutionTool','true')
    
    

    mvs = ET.SubElement(fmd,'ModelVariables')
    
    variable_index = 0

    for var in fmu_instance.vars:

        vref = str(var.value_reference)
        v = var.variability.value
        c = var.causality.value
        t = var.data_type.value

        idx_comment = ET.Comment(f'Index of variable = "{variable_index + 1}"')
        mvs.append(idx_comment)
        sv = ET.SubElement(mvs, "ScalarVariable")
        sv.set("name",var.name)
        sv.set("valueReference",vref)
        sv.set("variability", v)
        sv.set("causality", c)

        if(var.description):
            sv.set("description",var.description)
        

        if(var.initial):
            i = var.initial.value
            sv.set('initial', i)
        
        
        val = ET.SubElement(sv, t)

        if(var.start is not None):
            s = str(var.start)
            val.set("start", s)
        
        variable_index += 1


    ms = ET.SubElement(fmd,'ModelStructure')
    
    

    # 2.2.8) For each output we must declare 'Outputs' and 'InitialUnknowns'
    outputs = [(idx+1,o) for idx,o in enumerate(fmu_instance.vars) if o.causality.name == Fmi2Causality.output.name]

    if(outputs):
        os = ET.SubElement(ms,'Outputs')
        for idx,o in outputs:
            ET.SubElement(os,'Unknown',{'index' : str(idx), 'dependencies' : ''})

        os = ET.SubElement(ms, 'InitialUnknowns')
        for idx,o in outputs:
            ET.SubElement(os,'Unknown',{'index' : str(idx), 'dependencies' : ''})
    

    try:
        stream = io.StringIO()
        ElementTree(fmd).write(stream, encoding="unicode", short_empty_elements=True,xml_declaration=True)
    except Exception as e:
        raise RuntimeError("Failed to parse model description. ") from e
   
    # Format XML
    md_not_formatted =  stream.getvalue()
    md_formatted = minidom.parseString(md_not_formatted).toprettyxml(indent='   ')


    return md_formatted
    






from os.path import join, exists, isdir, isfile
import os
import json
from tempfile import TemporaryDirectory
from pathlib import Path

from pyfmu.builder.generate import create_project


class TestGenerate():

    def test_returnsProject_pathsAreCorrect(self, tmpdir):

        tmpdir = Path(tmpdir)

        p = create_project(tmpdir, 'Adder')

        assert p.root.samefile(tmpdir)
        assert p.project_configuration_path.samefile(tmpdir / 'project.json')

    def test_generate_mainScriptTemplateAdded(self, tmpdir):

        p = create_project(tmpdir, 'Adder')

        assert p.main_script == 'adder.py'
        assert p.main_class == 'Adder'
        assert p.main_script_path == p.root / 'resources' / 'adder.py'


    def test_generate_configurationAdded(self, tmpdir):
        p = create_project(tmpdir, 'Adder')

        assert p.project_configuration_path.is_file()

        with open(p.project_configuration_path, 'r') as f:
            c = json.load(f)
            assert c['main_class'] == 'Adder'
            assert c['main_script'] == 'adder.py'



"""
This module contains functionality to the creation of Python FMU projects
"""
from os import makedirs
from os.path import basename, curdir, isdir, join, realpath, dirname, exists, normpath
from os.path import exists, isfile
from pathlib import Path
from shutil import copy, copytree, rmtree
import json

from jinja2 import Template

from pyfmu.resources import Resources




def _create_config(config_path: str, class_name: str, relative_script_path: str):
        
    with open(config_path, 'w') as f:
        json.dump({
            "main_script" : relative_script_path,
            "main_class": class_name
        }, f,indent=4)


def read_configuration(config_path : str) -> object:

    print(f"configuration path is {config_path}")
    if(not exists(config_path)):
        raise FileNotFoundError("Failed to read configuration, the file does not exist")

    if(not isfile(config_path)):
        raise FileNotFoundError("Failed to read configuration, the specified path does not point to a file")


    try:
        with open(config_path,'r') as f:
            config = json.load(f)
    except Exception as e:
        raise RuntimeError("Failed to parse project configuration file. Ensure that it is well formed json")
    
    # TODO use json schema
    isWellFormed = hasattr(config,'main_script') and hasattr(config,'class_name')

    return config
        


class PyfmuProject():
    """Object representing an pyfmu project.
    """

    def __init__(self,
                 root: Path,
                 main_class: str = None,
                 main_script: str = None,
                 main_script_path: Path = None,
                 project_configuration: dict = None,
                 project_configuration_path: Path = None,
                 pyfmu_dir: Path = None,
                 ):

        self.main_class = main_class
        self.main_script = main_script
        self.main_script_path = main_script_path
        self.project_configuration = project_configuration
        self.project_configuration_path = project_configuration_path
        self.pyfmu_dir = pyfmu_dir
        self.root = root

    @staticmethod
    def from_existing(p: Path):
        """Instantiates an object representation based on an existing project

        Arguments:
            p {Path} -- path to the root of the project
        """
        try:
            p = Path(p)

        except Exception:
            raise RuntimeError(
                'Unable to load project. The specified argument could not be converted to a path')

        if(not p.is_dir()):
            raise FileNotFoundError(
                f'Unable to load project. The specified path does not appear to be a directory: {p}')

        # Verify that path points to a valid pyfmu project.

        # 1. Should be directory
        if(not Path.is_dir):
            raise ValueError(
                'Specified path does not point to a directory, ensure that the path is correct.')

        # 2. Should contain a 'project.json' in the root
        has_project_json = (p / 'project.json').is_file()

        if(not has_project_json):
            raise ValueError(
                'The directory does not contain a "project.json" file.')

        with open(p/'project.json') as f:
            project_json = json.load(f)

        # 2+ TODO validate using json schema
        main_script = project_json['main_script']
        main_class = project_json['main_class']
        main_script_path = p / 'resources' / main_script
        project_configuration_path = p / 'project.json'
        project_configuration = project_json
        pyfmu_dir = p / 'resources' / 'pyfmu'

        # 3. Should contain resources folder
        has_resources = (p / 'resources').is_dir()

        if(not has_resources):
            raise ValueError(
                'The directory does not contain a resource folder.')

        # 4. main script should exist inside resources.
        has_main_script = (p / 'resources' / main_script).is_file()

        if(not has_main_script):
            raise ValueError(
                f'The main python script: {main_script} could not be found in the resources folder. Ensure that the "project.json" defines the correct script.')

        # 5. TODO main class should be defined by main script

        project = PyfmuProject(root=p,
                               main_script=main_script,
                               main_class=main_class,
                               main_script_path=main_script_path,
                               project_configuration=project_configuration,
                               project_configuration_path=project_configuration_path,
                               pyfmu_dir=pyfmu_dir
                               )

        return project


def _create_dirs(project_path: str, exist_ok: bool = True):
    resources_path = join(project_path, "resources")

    try:
        makedirs(resources_path, exist_ok=True)
    except OSError:
        raise Exception(
            "Project with identical name already exists in that location. Please specify a new name, another directory or remove the old project.")
    except Exception as e:
        raise Exception("Failed to create directories for the Python project")


def _generate_fmu_template(template_path: str, main_class_name: str, script_output_path: str) -> None:

    r = None
    with open(template_path, 'r') as f:
        s = f.read()
        template = Template(s)

        r = template.render(
            {
                "class_name": main_class_name,
                "description": "",
                "model_name": main_class_name,
                "author": ""
            }
        )

    makedirs(dirname(script_output_path), exist_ok=True)

    with open(script_output_path, 'w') as f:
        f.write(r)


def _copy_pyfmu_to_project(project: PyfmuProject) -> PyfmuProject:

    resource_pyfmu_dir = Resources.get().pyfmu_dir

    project_pyfmu_dir = project.root / 'resources' / 'pyfmu'

    try:
        copytree(resource_pyfmu_dir, project_pyfmu_dir)
    except Exception as e:
        raise RuntimeError(
            "pyfmu library exists but could not be copied the generated project.") from e

    project.pyfmu_dir = project_pyfmu_dir

    return project


def _write_templateScript_to_project(project: PyfmuProject):

    resource_scriptTemplate_path = Resources.get().scriptTemplate_path

    with open(resource_scriptTemplate_path, 'r') as f:
        s = f.read()
        template = Template(s)

        r = template.render(
            {
                "class_name": project.main_class,
                "description": "",
                "model_name": project.main_class,
                "author": ""
            }
        )

    project_scriptTemplate_path = project.root / 'resources' / project.main_script

    makedirs(project_scriptTemplate_path.parent, exist_ok=True)

    with open(project_scriptTemplate_path, 'w') as f:
        f.write(r)

    project.main_script_path = project_scriptTemplate_path


def _write_projectConfig_to_project(project: PyfmuProject):

    project_configuration_path = project.root / 'project.json'

    config = {
        "main_class": project.main_class,
        "main_script": project.main_script
    }

    with open(project_configuration_path, 'w') as f:
        json.dump(config, f)

    project.project_configuration = config
    project.project_configuration_path = project_configuration_path


def create_project(project_path: str, main_class_name: str, overwrite=True) -> PyfmuProject:
    """Creates a new PyFMU project at the specified path.

    Parameters
    ----------
    project_path : str
        output path of the project
    main_class_name : str
        name of the main script 
    overwrite : bool, optional
        if true overwrite any existing files at the specified output path, by default True

    Returns
    -------
    PyfmuProject
        An object representing the exported project. 
        This may be used to access the paths of the various project artifacts.
    """
    project_path = Path(project_path)

    if(overwrite and exists(project_path)):
        rmtree(project_path)

    # TODO validate script names
    main_script = main_class_name.lower() + '.py'
    project = PyfmuProject(
        project_path, main_class=main_class_name, main_script=main_script)

    #_copy_pyfmu_to_project(project)
    _write_templateScript_to_project(project)
    _write_projectConfig_to_project(project)

    return project



from pathlib import Path

import pytest

from pyfmu.builder.export import PyfmuProject, PyfmuArchive
from pyfmu.tests import get_example_project, ExampleArchive, get_all_examples


def get_empty_archive(root: Path) -> PyfmuArchive:
    return PyfmuArchive(root, "")


def get_empty_project(root: Path) -> PyfmuProject:
    return PyfmuProject(root, None, None)


def test_sourcesCopied():

    with ExampleArchive('Adder') as a:
        sources_copied = a.main_script_path.is_file()

        assert sources_copied


def test_binariesCopied():
    with ExampleArchive('Adder') as a:

        assert a.binaries_dir.is_dir()
        assert (a.binaries_dir / 'win64' / 'pyfmu.dll').is_file()
        assert (a.binaries_dir / 'linux64' / 'pyfmu.so').is_file()


def test_slaveConfigurationGenerated():
    with ExampleArchive('Adder') as a:
        slaveConfiguration_generated = a.slave_configuration_path.is_file()

        assert slaveConfiguration_generated


def test_modelDescriptionGenerated():

    with ExampleArchive('Adder') as a:
        assert a.model_description is not None
        assert a.model_description_path.is_file()


def test_export_multipleInRow_modelDescriptionCorrect():

    fnames = get_all_examples()
    mds = []
    for f in fnames:
        with ExampleArchive(f) as a:
            assert a.model_description_path.is_file()
            mds.append(a.model_description)


def test_fromExisting_projectExists_OK():

    p = get_example_project('Adder')

    project = PyfmuProject.from_existing(p)

    assert project.root == p
    assert project.main_class == 'Adder'
    assert project.main_script == 'adder.py'


def test_fromExisting_emptyDirectory_Throws(tmpdir):

    with pytest.raises(ValueError):
        PyfmuProject.from_existing(tmpdir)



from abc import ABC, abstractmethod
from typing import List, Iterable, Tuple
from uuid import uuid4
import logging

from pyfmu.fmi2 import Fmi2ScalarVariable,Fmi2LogMessage, Fmi2Logger,Fmi2Causality, Fmi2DataTypes, Fmi2Initial, Fmi2Variability, Fmi2Status


_internal_log_catergory = 'pyfmu'

log = logging.getLogger('fmu')


class Fmi2Slave:

    def __init__(self, modelName: str, author="", copyright="", version="", description="", standard_log_categories=True):
        """Constructs a FMI2

        Arguments:
            modelName {str} -- [description]

        Keyword Arguments:
            author {str} -- [description] (default: {""})
            copyright {str} -- [description] (default: {""})
            version {str} -- [description] (default: {""})
            description {str} -- [description] (default: {""})
            standard_log_categories {bool} -- registers standard logging categories defined by the FMI2 specification (default: {True})
        """

        self.author = author
        self.copyright = copyright
        self.description = description
        self.modelName = modelName
        self.license = license
        self.guid = uuid4()
        self.vars = []
        self.version = version
        self.value_reference_counter = 0
        self.used_value_references = {}

        self.logger = Fmi2Logger()
        if(standard_log_categories):
            self.logger.register_all_standard_categories()

    def register_variable(self,
                          name: str,
                          data_type: Fmi2DataTypes = None,
                          causality=Fmi2Causality.local,
                          variability=Fmi2Variability.continuous,
                          initial: Fmi2Initial = None,
                          start=None,
                          description: str = "",
                          define_attribute: bool = True,
                          value_reference: int = None
                          ):
        """Add a variable to the model such as an input, output or parameter.

        Arguments:
            name {str} -- [description]
            data_type {Fmi2DataTypes} -- [description]

        Keyword Arguments:
            causality {[type]} -- defines the type of the variable, such as input, output or parameter (default: {Fmi2Causality.local})
            variability {[type]} -- defines the time dependency of the variable. (default: {Fmi2Variability.continuous})
            initial {Fmi2Initial} -- defines how the variable is initialized (default: {None})
            start {[type]} -- start value of the variable. (default: {None})
            description {str} -- a description of the variable which is added to the model description (default: {""})
            define_attribute {bool} -- if true, automatically add the specified attribute to instance if it does not already exist. (default: {True})

        Inference Rules:
            i1. shorthands and aliases:
                data_type,causality,initial and variablity can be defined using string shorthands.

            i2. default start values:
                if data_type is defined but start is undefined, default values are used based on data_type 

            i3. data type inference:
                if data_type is undefined but start is defined, data_type is inferred from start's type.

            i4. initial based on causality and variability (2.2.7 p.49)
                Default initial are defined based on the combination of causality and variability:
                    * exact : constant[local|output], [fixed|tunable]parameter
                    * calculated : [fixed|tunable][calculatedParameter|local], [continous|discrete][local|output]
                    * (do not define) : [discrete|continuous]input, continous independent.


        Validation Rules:
            v1: data type and causality (2.2.7 p.48)
                only real valued variables may have causality continuous

            v2: causality and variablity (2.2.7 p.48-49)
                Only the following combinations of variability and causality are allowed:
                * calculatedParameter: {fixed,tunable}
                * independent: {continuous}
                * input : {continuous,discrete}
                * local : {continuous,constant,discrete,fixed,tuneable}
                * output : {continuous,constant,discrete}
                * parameter : {fixed,tuneable}

            v3: initial allowed (2.2.7 p.49)
                Only certain certain intials are allowed for a given combination of varability and causality.

            v4: should define start (2.2.7 p.47)
                Only certain combinations of causality, intial are allowed to provide start values.
                * [exact|approx] -> must define start
                * calculated -> must not define start
                * 

            Ordering:
            i1 -> i2 -> i3 -> v2 -> i4 -> v1 -> v3

        """

        #i1. shorthands and aliases
        type_aliases = {

            None: None,

            Fmi2DataTypes.real: Fmi2DataTypes.real,
            'real': Fmi2DataTypes.real,
            float: Fmi2DataTypes.real,

            Fmi2DataTypes.boolean: Fmi2DataTypes.boolean,
            'bool': Fmi2DataTypes.boolean,
            'boolean': Fmi2DataTypes.boolean,
            bool: Fmi2DataTypes.boolean,

            Fmi2DataTypes.integer: Fmi2DataTypes.integer,
            'int': Fmi2DataTypes.integer,
            'integer': Fmi2DataTypes.integer,
            int: Fmi2DataTypes.integer,

            Fmi2DataTypes.string: Fmi2DataTypes.string,
            'string': Fmi2DataTypes.string,
            'str': Fmi2DataTypes.string,
            str: Fmi2DataTypes.string
        }
        causality_aliases = {
            None: None,

            Fmi2Causality.parameter: Fmi2Causality.parameter,
            'parameter': Fmi2Causality.parameter,

            Fmi2Causality.calculatedParameter: Fmi2Causality.calculatedParameter,
            'calculatedparameter': Fmi2Causality.calculatedParameter,

            Fmi2Causality.input: Fmi2Causality.input,
            'input': Fmi2Causality.input,

            Fmi2Causality.output: Fmi2Causality.output,
            'output': Fmi2Causality.output,

            Fmi2Causality.local: Fmi2Causality.local,
            'local': Fmi2Causality.local,

            Fmi2Causality.independent : Fmi2Causality.independent,
            'independent': Fmi2Causality.independent,
        }
        initial_aliases = {
            None: None,
            Fmi2Initial.exact: Fmi2Initial.exact,
            'exact': Fmi2Initial.exact,

            Fmi2Initial.approx: Fmi2Initial.approx,
            'approx': Fmi2Initial.approx,

            Fmi2Initial.calculated: Fmi2Initial.calculated,
            'calculated': Fmi2Initial.calculated
        }
        variability_aliases = {
            None: None,
            Fmi2Variability.constant: Fmi2Variability.constant,
            'constant': Fmi2Variability.constant,

            Fmi2Variability.fixed: Fmi2Variability.fixed,
            'fixed': Fmi2Variability.fixed,

            Fmi2Variability.tunable: Fmi2Variability.tunable,
            'tunable': Fmi2Variability.tunable,

            Fmi2Variability.discrete: Fmi2Variability.discrete,
            'discrete': Fmi2Variability.discrete,

            Fmi2Variability.continuous: Fmi2Variability.continuous,
            'continuous': Fmi2Variability.continuous
        }

        if(data_type not in type_aliases):
            raise ValueError(
                f'Unrecognized data type: {data_type}. Possible values are {type_aliases.keys()}')

        if(causality not in causality_aliases):
            raise ValueError(
                f'Unrecognized causality: {causality}. Possible values are {causality_aliases.keys()}')

        if(initial not in initial_aliases):
            raise ValueError(
                f'Unrecognized initial: {initial}. Possible values are {initial_aliases.keys()}')

        if(variability not in variability_aliases):
            raise ValueError(
                f'Unrecognized initial: {variability}. Possible values are {variability_aliases.keys()}')

        data_type = type_aliases[data_type]
        causality = causality_aliases[causality]
        initial = initial_aliases[initial]
        variability = variability_aliases[variability]

        # i4. intial default

        case_a = (Fmi2Initial.exact, {Fmi2Initial.exact})
        case_b = (Fmi2Initial.calculated, {
                  Fmi2Initial.approx, Fmi2Initial.calculated})
        case_c = (Fmi2Initial.calculated, {
                  Fmi2Initial.approx, Fmi2Initial.calculated, Fmi2Initial.exact})
        case_de = (None, {None})

        variabilityAndCausality_to_intial = {
            (Fmi2Variability.constant, Fmi2Causality.local): case_a,
            (Fmi2Variability.constant, Fmi2Causality.output): case_a,
            (Fmi2Variability.fixed, Fmi2Causality.parameter) : case_a,
            (Fmi2Variability.tunable, Fmi2Causality.parameter) : case_a,

            (Fmi2Variability.fixed, Fmi2Causality.calculatedParameter): case_b,
            (Fmi2Variability.fixed, Fmi2Causality.local): case_b,
            (Fmi2Variability.tunable, Fmi2Causality.calculatedParameter): case_b,
            (Fmi2Variability.tunable, Fmi2Causality.local): case_b,
      
    
            (Fmi2Variability.discrete, Fmi2Causality.output): case_c,
            (Fmi2Variability.discrete, Fmi2Causality.local): case_c,
            (Fmi2Variability.continuous, Fmi2Causality.output): case_c,
            (Fmi2Variability.continuous, Fmi2Causality.local): case_c,

            (Fmi2Variability.discrete, Fmi2Causality.input) : case_de,
            (Fmi2Variability.continuous, Fmi2Causality.input) : case_de,
            (Fmi2Variability.continuous, Fmi2Causality.independent) : case_de,
        }

        if(initial is None):
            initial = variabilityAndCausality_to_intial[(
                variability, causality)][0]

        # i2. default start values + v4. should define start
        must_define_start = (initial in {Fmi2Initial.exact, Fmi2Initial.approx}
                             or causality in {Fmi2Causality.parameter, Fmi2Causality.input}
                             or variability in {Fmi2Variability.constant})

        can_not_define_start = (
            initial == Fmi2Initial.calculated or causality == Fmi2Causality.independent)

        assert(must_define_start != can_not_define_start)

        if(must_define_start and start is None and data_type is not None):

            type_to_start = {
                Fmi2DataTypes.boolean: False,
                Fmi2DataTypes.integer: 0,
                Fmi2DataTypes.real: 0.0,
                Fmi2DataTypes.string: ''
            }
            start = type_to_start[data_type]

        elif (must_define_start and start is None and data_type is None):
            raise ValueError(
                f"""A start value must be specified for the combination of causality : {causality}, intial : {initial} and variability {variability}.
                 Specify either a start value or the datatype such that a default will be provided.""")

        elif (not must_define_start and start is not None):
            raise ValueError(
                f"""Start value must NOT be specified for the combination of causality : {causality}, intial : {initial} and variability {variability}.""")

        # i3. data type inference
        start_to_type = {
            bool: Fmi2DataTypes.boolean,
            int: Fmi2DataTypes.integer,
            float: Fmi2DataTypes.real,
            str: Fmi2DataTypes.string
        }
        if(data_type is None and start is not None):
            for t in start_to_type:
                if(isinstance(start, t)):
                    data_type = start_to_type[t]
                    break

        #v2. causality and variablity
        if((variability, causality) not in variabilityAndCausality_to_intial):
            raise ValueError(
                f'Illegal combination of causality : {causality} and variablity : {variability}. The combination is not permitted.')


        #v1. type and causality
        if(data_type is not Fmi2DataTypes.real and variability is Fmi2Variability.continuous):
            raise ValueError(
                f'Illegal combination of type : {data_type} and variability : {variability}. Only real valued variables are allowed to be continuous')

        # v3 initial allowed
        if(initial not in variabilityAndCausality_to_intial[variability, causality][1]):
            raise ValueError(
                f'Illegal initial value : {initial} for combination of causality : {causality} and variability : {variability}')

        # if not specified find an unused value reference
        if(value_reference is None):
            value_reference = self._acquire_unused_value_reference()

        var = Fmi2ScalarVariable(name=name, data_type=data_type, initial=initial, causality=causality,
                             variability=variability, description=description, start=start, value_reference=value_reference)

        self.vars.append(var)

        if(define_attribute):
            self._define_variable(var)

    def register_log_category(self, name: str):
        """Registers a new log category.
        This information is used by co-simulation engines to filter messages

        Arguments:
            name {str} -- name of the category.

        Examples:

        ```
        self.register_log_category('runtime validation')

        ```
        """

        pass

    def setup_experiment(self, start_time: float):
        pass

    def enter_initialization_mode(self):
        pass

    def exit_initialization_mode(self):
        pass

    def do_step(self, current_time: float, step_size: float) -> bool:
        pass

    def reset(self):
        pass

    def terminate(self):
        pass

    def __set_debug_logging__(self, logging_on: bool, categories: Iterable[str]) -> None:
        """Defines the set of active log categories for which log messages will logged.
        Messages logged to any other categories will be ignored.


        This function is called by the tool though the fmi2SetDebugLogging function.

        Note that the tool can only "see" categories registered in the model description using the register_log_category.


        Arguments:
            categories {Iterable[str]} -- a set of categories which are set to be active
            logging_on {bool} -- if true, enable the categories, otherwise disable them

        Defaults category mapping:
            FMI2 specifies several standardized categories.

            * logEvents                 : Log all events (during initialization and simulation).
            * logSingularLinearSystems  : Log the solution of linear systems of equations if the solution is singular
                                          (and the tool picked one solution of the infinitely many solutions).
            * logNonlinearSystems       : Log the solution of nonlinear systems of equations.
            * logDynamicStateSelection  : Log the dynamic selection of states.
            * logStatusWarning          : Log messages when returning fmi2Warning status from any function.
            * logStatusDiscard          : Log messages when returning fmi2Discard status from any function.
            * logStatusError            : Log messages when returning fmi2Error status from any function.
            * logStatusFatal            : Log messages when returning fmi2Fatal status from any function.
            * logStatusPending          : Log messages when returning fmi2Pending status from any function.
            * logAll                    : Log all messages.

            The standard standard leaves it up to the implementation to decide which "categories" map to each of the log-categories.

            ``` Python
            self.log(Fmi2Status.OK,"events","initialized FMU")
            ```

        Examples:
        ``` Python
            # standard logging
            fmu = MyFMU()
            fmu.standard_log_catgories()
            fmu.__set_debug_logging__({'LogAll'})

            fmu.log(Fmi2Status.ok,'logEvents',)

        ```
        """

        self.logger.set_active_log_categories(logging_on, categories)

    def log(self, message: str, category=None, status=Fmi2Status.ok) -> None:
        """Logs a message to the fmi interface.

        Note that only categories which are set as active using the __set_debug_logging__, are propageted to the tool.
        The function is called by the tool using with the categories which the user wishes to be active.

        Arguments:
            status {Fmi2Status} -- The current status of the FMU.
            category {str} -- The category of the log message.
            message {str} -- The log message itself.

        Logging categories mappings:


        """

        self.logger.log(message, category, status)

    def __pop_log_messages__(self, n: int) -> Tuple[str, str, str]:
        """Function called by the wrapper to fetch log messages

        Arguments:
            n {int} -- Number of messages to fetch
        """
        if(n > len(self.logger)):
            self.log(
                f"Unable to pop messages. Requested number of log messages: {n}, is larger than the number currently available: {len(self.logger)}.")
            return None

        messages = self.logger.pop_messages(n)

        # for convenience we convert the object into tuples
        messages_tuples = [(m.status.value, m.category, m.message)
                           for m in messages]

        return messages_tuples

    def __get_log_size__(self) -> int:
        """Returns the number of log messages that are currently on the log stack.

        Returns:
            int -- [description]
        """
        return len(self.logger)

    def __get_integer__(self, vrs, refs):
        for i in range(len(vrs)):
            vr = vrs[i]
            var = self.vars[vr]
            if var.is_integer():
                refs[i] = getattr(self, var.name)
            else:
                raise Exception(
                    f"Variable with valueReference={vr} is not of type Integer!")

    def __get_real__(self, vrs, refs):
        for i in range(len(vrs)):
            vr = vrs[i]
            var = self.vars[vr]
            if var.is_real():
                refs[i] = getattr(self, var.name)
            else:
                raise Exception(
                    f"Variable with valueReference={vr} is not of type Real!")

    def __get_boolean__(self, vrs, refs):
        for i in range(len(vrs)):
            vr = vrs[i]
            var = self.vars[vr]
            if var.is_boolean():
                refs[i] = getattr(self, var.name)
            else:
                raise Exception(
                    f"Variable with valueReference={vr} is not of type Boolean!")

    def __get_string__(self, vrs, refs):
        for i in range(len(vrs)):
            vr = vrs[i]
            var = self.vars[vr]
            if var.is_string():
                refs[i] = getattr(self, var.name)
            else:
                raise Exception(
                    f"Variable with valueReference={vr} is not of type String!")

    def __set_integer__(self, vrs, values):
        for i in range(len(vrs)):
            vr = vrs[i]
            var = self.vars[vr]
            if var.is_integer:
                setattr(self, var.name, values[i])
            else:
                raise Exception(
                    f"Variable with valueReference={vr} is not of type Integer!")

    def __set_real__(self, vrs, values):
        for i in range(len(vrs)):
            vr = vrs[i]
            var = self.vars[vr]
            if var.is_real():
                setattr(self, var.name, values[i])
            else:
                raise Exception(
                    f"Variable with valueReference={vr} is not of type Real!")

    def __set_boolean__(self, vrs, values):
        for i in range(len(vrs)):
            vr = vrs[i]
            var = self.vars[vr]
            if var.is_boolean():
                setattr(self, var.name, values[i])
            else:
                raise Exception(
                    f"Variable with valueReference={vr} is not of type Boolean!")

    def __set_string__(self, vrs, values):
        for i in range(len(vrs)):
            vr = vrs[i]
            var = self.vars[vr]
            if var.is_string():
                setattr(self, var.name, values[i])
            else:
                raise Exception(
                    f"Variable with valueReference={vr} is not of type String!")

    def _define_variable(self, sv: Fmi2ScalarVariable):

        if(not hasattr(self, sv.name)):
            log.debug(f'adding')

            setattr(self, sv.name, sv.start)
            return

        if(sv.initial in {Fmi2Initial.exact, Fmi2Initial.approx}):
            old = getattr(self, sv.name)
            new = sv.start

            if(old != new):
                log.warning(
                    "start value variable defined using the 'register_variable' function does not match initial value")
                setattr(self, sv.name, new)

    def _acquire_unused_value_reference(self) -> int:
        """ Returns the an unused value reference
        """
        while(True):
            vr = self.value_reference_counter
            self.value_reference_counter += 1

            if(vr not in self.used_value_references):
                return vr



import subprocess
from enum import Enum
import platform
import logging
from tempfile import mkdtemp
import os
from os.path import realpath, join
from pathlib import Path

from pyfmu.resources import Resources
from pyfmu.builder import cd, has_java, has_fmpy


class FMI_Versions(Enum):
    FMI2 = "fmi2"
    FMI3 = "fmi3"


class ValidationResult():
    """
    Represents the result of validating an FMU with different validation techniques.
    """

    def __init__(self):
        self.validation_tools = {}

    def set_result_for(self, tool: str, valid: bool, message=""):
        self.validation_tools[tool] = {"valid": valid, "message": message}

    def get_result_for(self, tool: str):
        return self.validation_tools[tool]

    @property
    def valid(self):
        ss = {s['valid'] for s in self.validation_tools.values()}
        return False not in ss


def validate(fmu_archive: str, use_fmpy: bool = True, use_fmucheck: bool = False, use_vdmcheck: bool = False):

    if(use_fmpy and not has_fmpy()):
        raise ImportError(
            "Cannot validate exported module using fmpy, the module could not be loaded. Ensure that the package is installed.")

    valid = True

    if(use_fmpy):
        from fmpy import dump, simulate_fmu

        dump(fmu_archive)

        try:
            res = simulate_fmu(fmu_archive)
        except Exception as e:
            valid = False
            return repr(e)

    return None


def validate_project(project) -> bool:
    """Validate a project to ensure that it is consistent.

    Arguments:
        project {PyfmuProject} -- The project that is validated

    Returns:
        bool -- [description]
    """
    # TODO add validation
    return True


def validate_modelDescription(modelDescription: str, use_fmucheck=False, use_vdmcheck=False, vdmcheck_version=FMI_Versions.FMI2) -> ValidationResult:

    if(True not in {use_fmucheck, use_vdmcheck}):
        raise ValueError(
            "arguments must specifiy at least one verification tool")

    results = ValidationResult()

    if(use_vdmcheck):
        _validate_vdmcheck(modelDescription, results, vdmcheck_version)

    if(use_fmucheck):
        raise NotImplementedError()

    return results


def _validate_vdmcheck(modelDescription: str, validation_results: ValidationResult, fmi_version=FMI_Versions.FMI2):
    """Validate the model description using the VDMCheck tool.

    Arguments:
        modelDescription {str} -- textual representation of the model description.

    Keyword Arguments:
        fmi_version {FMI_Versions} -- [description] (default: {FMI_Versions.FMI2})

    Raises:
        ValueError: Raised if an the fmi_version is unknown or if the tool does not support validation thereof.

    Notes:
        VDMCheck is implemented in Java, as such it requires java to be available in the path.
    """

    if(not has_java()):
        raise RuntimeError(
            'Unable to perform validation using VDMCheck, java was not found in the systems path.')

    fmi_to_jar = {
        fmi_version.FMI2: Resources.get().VDMCheck2_jar,
        fmi_version.FMI3: Resources.get().VDMCheck3_jar
    }

    if(fmi_version not in fmi_to_jar):
        raise RuntimeError(
            f'Unable to perform validation using VDMCheck. Unsupported FMI version: {fmi_version}, supported versions are: {fmi_to_jar.keys()}')

    jar_path = str(fmi_to_jar[fmi_version].resolve())
    # Run VDMCheck
    result = subprocess.run(
        ['java', '-jar', jar_path, '-x', modelDescription], capture_output=True)

    def _vdmcheck_no_errors(results):
        stdout_contains_no_error = b'no errors found' in results.stdout.lower()
        stderr_is_empty = results.stderr == b''
        error_code_ok = results.returncode == 0
        return stdout_contains_no_error and stderr_is_empty and error_code_ok

    # convert output
    isValid = _vdmcheck_no_errors(result)
    validation_results.set_result_for('vdmcheck', isValid, result.stdout)



"""Export all example projects as FMUs to folder named exported
"""


from pathlib import Path

import tqdm

from pyfmu.tests import get_example_project, get_all_examples, get_example_directory

from pyfmu.builder.export import export_project
from pyfmu.builder.generate import PyfmuProject


def export_all():
    for name in get_all_examples():

        p = get_example_project(name)

        outdir = get_example_directory().parent / 'exported' / name
        project = PyfmuProject.from_existing(p)
        export_project(project, outdir, overwrite=True)


if __name__ == "__main__":

    export_all()



from distutils.dir_util import copy_tree
from logging import debug, info, warning, error
from os import makedirs, listdir, rename
from os.path import join, curdir, dirname, splitext, basename, isdir, realpath, normpath, exists, splitext, relpath
from pathlib import Path
from shutil import copyfile, copytree, rmtree, make_archive, move, copy
import importlib
import json
import json
import logging
import sys

from pyfmu.builder import read_configuration,extract_model_description_v2,validate_project,PyfmuProject
from pyfmu.resources import Resources

_log = logging.getLogger(__name__)

_exists_ok = True






class PyfmuArchive():
    """Object representation of exported Python FMU.
    """

    def __init__(self,
                 root: Path = None,
                 resources_dir: Path =  None,
                 slave_configuration_path : Path = None,
                 binaries_dir : Path = None,
                 wrapper_win64 : Path = None,
                 wrapper_linux64: Path = None,
                 main_script_path : Path = None,
                 model_description: Path = None,
                 model_description_path : Path = None,
                 main_script: Path = None,
                 main_class : Path = None,
                 pyfmu_dir : Path = None
    ):
        """Creates an object representation of the exported Python FMU.

        Arguments:
            model_description {str} -- The model description of the exported FMU.
        """
        self.model_description = model_description

        self.main_script = main_script
        self.main_class = main_class
        self.slave_configuration = None

        # paths
        self.root = root
        self.resources_dir = resources_dir
        self.slave_configuration_path = slave_configuration_path
        self.main_script_path = main_script_path
        self.model_description_path = model_description_path
        self.binaries_dir = binaries_dir
        self.wrapper_win64 = wrapper_win64
        self.wrapper_linux64 = wrapper_linux64
        self.pyfmu_dir = pyfmu_dir


def import_by_source(path: str):
    """Loads a python module using its name and the path to the python source script.

    Arguments:
        path {str} -- path to the module

    Returns:
        module -- module loaded from the source file.
    """

    module = splitext(basename(path))[0]

    sys.path.append(dirname(path))

    spec = importlib.util.spec_from_file_location(module, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    sys.path.pop()

    return module


def _available_export_platforms():
    return ["Win64", "Linux64"]


def _create_archive_directories(archive_path, exist_ok=True):

    archive_resources_path = join(archive_path, "resources")

    platforms = _available_export_platforms()

    binary_dirs = [join(archive_path, "binaries", p) for p in platforms]
    binary_dirs = []

    try:
        makedirs(archive_resources_path, exist_ok=exist_ok)

        for d in binary_dirs:
            makedirs(d, exist_ok=exist_ok)

    except Exception as e:
        raise RuntimeError("Failed to create archive directories")


def _resources_to_archive(project_dir, builder_resources_dir, archive_dir):

    # TODO ensure that binaries actually exist

    # copy binaries / python wrapper into archive
    builder_binaries_dir = join(builder_resources_dir, "wrapper", "binaries")
    archive_binaries_dir = join(archive_dir, "binaries")
    copytree(builder_binaries_dir, archive_binaries_dir)

    # copy source files into archive
    project_resources_dir = join(project_dir, "resources")
    archive_resources_dir = join(archive_dir, "resources")
    copytree(project_resources_dir, archive_resources_dir)


def _copy_pyfmu_lib_to_archive(archive: PyfmuArchive, project: PyfmuProject = None) -> PyfmuArchive:
    """ Copies the Python library into the archive.

    If a project is not specified a 'clean' copy from the resources folder is used. 
    Otherwise, if a project is specified the library is copied from this folder instead.

    Arguments:
        archive {PyfmuArchive} -- The FMU

    Keyword Arguments:
        project {PyfmuProject} -- If specified, the potentially modified pyfmu library instead of one from the resources. (default: {None})
    """

    copy_from_resources = project is None

    if(copy_from_resources):
        
        lib_path = Resources.get().pyfmu_dir
        err_msg = 'No directory named pyfmu was found in the resources directory, likely this has been deleted.'

    else:
        lib_path = project.root / 'resources' / 'pyfmu'
        err_msg = "No directory named pyfmu was found in the projects resources. If the intention is to copy the 'local' library, ensure that is is present in the resources folder of the project. Otherwise, to use a clean copy omit specifying the project."

    lib_exists = lib_path.is_dir()

    if(not lib_exists):
        raise RuntimeError(
            f'Failed to copy the library to the archive. {err_msg}')

    archive_lib_dir = archive.root / 'resources' / 'pyfmu'

    copytree(lib_path, archive_lib_dir)

    archive.pyfmu_dir = archive_lib_dir

    return archive

def _copy_binaries_to_archive(archive: PyfmuArchive) -> PyfmuArchive:
    """Copies the binaries to the archive.
    
    Arguments:
        archive {PyfmuArchive} -- The archive being exported
    
    Returns:
        PyfmuArchive -- The archive being exported
    """

    binaries_path = Resources.get().binaries_dir


    archive_binaries_path = archive.root / 'binaries'

    copytree(binaries_path,archive_binaries_path)

    # paths
    archive.binaries_dir = archive_binaries_path
    archive.wrapper_win64 = archive.binaries_dir / 'win64' / 'pyfmu.dll'
    archive.wrapper_linux64 = archive.binaries_dir / 'linux64' / 'pyfmu.so'

    return archive

def _copy_sources_to_archive(project: PyfmuProject, archive: PyfmuArchive) -> PyfmuArchive:
    """Copies the source files of the project into the archive.

    Note that that a distinction is made between source files and the pyfmu library. 

    In this case source files refer to files written by the developer.

    NOTE: Currently this only copies main script.

    Arguments:
        project {PyfmuProject} -- [description]
        archive {PyfmuArchive} -- [description]
    """

    main_script_found = project.main_script_path.is_file()

    if(not main_script_found):
        raise RuntimeError(
            f'main script: {project.main_script} was not found inside project: {project.root}')

    archive_main_script_path = archive.root / 'resources' / archive.main_script

    # make directories and copy source files

    if(not archive_main_script_path.parent.is_dir()):
        makedirs(archive_main_script_path.parent)
    
    copyfile(project.main_script_path,archive_main_script_path)

    archive.main_script_path = archive_main_script_path

def _write_modelDescription_to_archive(project : PyfmuProject, archive : PyfmuArchive) -> PyfmuArchive:
    
    instance = _instantiate_main_class(archive.main_script_path, archive.main_class)
    md = extract_model_description_v2(instance)

    archive_model_description_path = archive.root / 'modelDescription.xml'
    
    with open(archive_model_description_path,'w') as f:
        f.write(md)

    archive.model_description = md
    archive.model_description_path = archive_model_description_path

def _write_slaveConfiguration_to_archive(project : PyfmuProject, archive : PyfmuArchive) -> PyfmuArchive:
    
    archive_slave_configuration_path = archive.root / 'resources' / 'slave_configuration.json'

    # currently project and slave configuration contains same info
    slave_configuration  = project.project_configuration

    makedirs(archive_slave_configuration_path.parent)

    with open(archive_slave_configuration_path,'w') as f:
        json.dump(slave_configuration,f)

    archive.slave_configuration_path = archive_slave_configuration_path
    archive.slave_configuration = slave_configuration
    archive.main_class = slave_configuration['main_class']
    archive.main_script = slave_configuration['main_script']

    return PyfmuArchive



def _compress(archive_path: str):
    extension = "zip"
    make_archive(archive_path, 'zip', archive_path)
    rename(f"{archive_path}.{extension}", f"{archive_path}.fmu")

def _generate_model_description(main_script_path: str, main_class: str) -> str:

    instance = _instantiate_main_class(main_script_path, main_class)
    md = extract_model_description_v2(instance)

    return md


def _generate_slave_config(project: PyfmuProject):
    """Generate a slave configuration file for the project

    Note: Currently there is no difference between the project.json and slave_configuration.json files.
    The difference 

    Arguments:
        project {PyfmuProject} -- [description]

    Returns:
        [type] -- Dict
    """
    return project.project_configuration


def _instantiate_main_class(main_script_path: str, main_class: str):
    module = import_by_source(main_script_path)

    main_class_ctor = getattr(module, main_class)

    if(main_class_ctor is None or not callable(main_class_ctor)):
        raise RuntimeError(
            f"Failed to generate model description. The specified file {main_script_path} does not define any callable attribute named {main_class}.")

    try:
        main_class_instance = main_class_ctor()
    except Exception as e:
        raise RuntimeError(
            f"Failed generating model description, The construtor of the main class threw an exception. Ensure that the script defines a parameterless constructor. Error message was: {repr(e)}") from e

    return main_class_instance


def _validate_model_description(md: str) -> bool:
    return True


def export_project(project: PyfmuProject, outputPath: Path, overwrite=False, store_compressed : bool = True) -> PyfmuArchive:
    """Exports a pyfmu project as an FMU to the specified output path.
    
    Arguments:
        project {PyfmuProject} -- A PyfmuProject or the path to a project. 
        outputPath {Path} -- The path to which the FMU is exported.
    
    Keyword Arguments:
        store_compressed {bool} -- compress the exported FMU (default: {True})
    
    Raises:
        RuntimeError: [description]
    
    Returns:
        PyfmuArchive -- An object representing the exported FMU.
    

    """
    isProject = isinstance(project, PyfmuProject)
    if(not isProject):

        try:
            project = PyfmuProject.from_existing(Path(project))
        except Exception as e:
            raise ValueError('Could not load the specified project. The path project argument appears to be neither a project or the path to a project') from e
        
    try:
        outputPath = Path(outputPath)
    except Exception as e:
        raise(ValueError('The output path does not appear to be a valid path, please ensure that it is correct.'))

    is_valid_project = validate_project(project)

    if(not is_valid_project):
        raise RuntimeError(
            'The specified project can not be exported due to errors in the project.')

    if(overwrite and exists(outputPath)):
        rmtree(outputPath, ignore_errors=True)

    # create object representation of the project being exported
    # The purpose of this is to encapsulate path related stuff
    archive = PyfmuArchive(
        
        binaries_dir = outputPath / 'binaries',
        model_description_path = outputPath / 'modelDescription.xml',
        root=outputPath,
        resources_dir = outputPath / 'resources',
        slave_configuration_path = outputPath / 'resources' / 'slave_configuration.json'
    )

    # read slave configuration
    _write_slaveConfiguration_to_archive(project,archive)

    # copy source files to archive
    _copy_sources_to_archive(project, archive)

    _copy_binaries_to_archive(archive)

    # generate model description and write
    _write_modelDescription_to_archive(project,archive)


       
    return archive




from abc import ABC

from pyfmu.fmi2 import validate_vc, get_default_initial, get_possible_initial,Fmi2DataTypes, Fmi2Initial, Fmi2Causality, Fmi2Variability


class Fmi2ScalarVariable(ABC):

    def __init__(self,
                 name: str, 
                 data_type: Fmi2DataTypes,
                 initial: Fmi2Initial = None,
                 causality=Fmi2Causality.local,
                 variability=Fmi2Variability.continuous,
                 start = None,
                 description: str = "",
                 value_reference: int = None):

        err = validate_vc(
            variability, causality)

        if(err is not None):
            raise Exception(
                "Illegal combination fo variability and causality, FMI2 specification describes the issue with this combination as:\n" + err)

        initial = initial if initial is not None else get_default_initial(
            variability, causality)

        allowed_initial = get_possible_initial(variability, causality)

        is_valid_initial = initial in allowed_initial

        if(not is_valid_initial):
            raise Exception(
                "Illegal combination of variabilty causality, see FMI2 spec p.49 for legal combinations")


        is_valid_start = Fmi2ScalarVariable.validate_start_value(
            data_type=data_type,
            causality=causality,
            variability = variability,
            initial=initial,
            start=start)

        if(is_valid_start != None):
            raise Exception("Illegal start value\n")

        self.causality = causality
        self.data_type = data_type
        self.description = description
        self.initial = initial
        self.name = name
        self.variability = variability
        self.start = start
        self.value_reference = value_reference
        

    def is_real(self) -> bool:
        return self.data_type == Fmi2DataTypes.real

    def is_integer(self) -> bool:
        return self.data_type == Fmi2DataTypes.integer

    def is_boolean(self) -> bool:
        return self.data_type == Fmi2DataTypes.boolean

    def is_string(self) -> bool:
        return self.data_type == Fmi2DataTypes.string

    @staticmethod
    def validate_start_value(data_type, causality, initial, variability, start):
        must_be_defined = Fmi2ScalarVariable.should_define_start(variability, causality, initial)

        is_defined = start != None
        if(must_be_defined ^ is_defined):
            s = "must be defined" if not is_defined else "may not be defined"
            return f"Start values {s} for this combination of variability: {variability}, causality: {causality} and intial: {initial}"


        if(must_be_defined):

            allowed_types = {
                Fmi2DataTypes.real : ({int,float},int.__init__),
                Fmi2DataTypes.boolean : ({bool},bool.__init__),
                Fmi2DataTypes.integer : ({int},int.__init__),
                Fmi2DataTypes.string : ({str},str.__init__),
            }


            types,_ = allowed_types[data_type]
        
            valid_conversion = any([isinstance(start,t) for t in types])
            
            if(not valid_conversion):        
                raise TypeError(f'Illegal combination of data type and start value. Type is {data_type} start is {start}.')
        

        return None

    @staticmethod
    def should_define_start(variability: Fmi2Variability, causality: Fmi2Causality, initial: Fmi2Initial) -> bool:
        """Returns true if the combination requires that a start value is defined, otherwise false.

        For reference check the FMI2 specification p.54 for a description of which combination are allowed.
        """
        # see fmi2 spec p.54
        must_define_start = (initial in {Fmi2Initial.exact, Fmi2Initial.approx}
                            or causality in {Fmi2Causality.parameter, Fmi2Causality.input}
                            or variability in {Fmi2Variability.constant})

        can_not_define_start = (
            initial == Fmi2Initial.calculated or causality == Fmi2Causality.independent)

        # should be mutually exclusive
        assert(must_define_start != can_not_define_start)

        return must_define_start

    def __repr__(self):
        initial_str = self.initial.value if self.initial else 'notPermitted'
        return f'{self.name}:{self.data_type.value}:{self.causality.value}:{initial_str}'

    def __str__(self):
        return self.__repr__()



from collections import defaultdict
from typing import Union, Any

from pyfmu.fmi2 import Fmi2Variability, Fmi2Causality, Fmi2Initial

# dictionaries defining valid combinations of variability and causality, see fmi2 p.49
_vc_combinations = defaultdict(dict)


_a_error = """The combinations “constant / parameter”, “constant / calculatedParameter” and “constant /
input” do not make sense, since parameters and inputs are set from the environment,
whereas a constant has always a value.
"""

_b_error = """The combinations “discrete / parameter”, “discrete / calculatedParameter”, “continuous /
parameter” and continuous / calculatedParameter do not make sense, since causality =
“parameter” and “calculatedParameter” define variables that do not depend on time, whereas
“discrete” and “continuous” define variables where the values can change during simulation.
"""

_c_error = "For an “independent” variable only variability = “continuous” makes sense."

_d_error = """A fixed or tunable “input” has exactly the same properties as a fixed or tunable parameter.
For simplicity, only fixed and tunable parameters shall be defined.
"""

_e_error = """A fixed or tunable “output” has exactly the same properties as a fixed or tunable
calculatedParameter. For simplicity, only fixed and tunable calculatedParameters shall be
defined.
"""

_A_initial = {"default": Fmi2Initial.exact, "possible": {Fmi2Initial.exact}}
_B_initial = {"default": Fmi2Initial.calculated,
              "possible": {Fmi2Initial.approx, Fmi2Initial.calculated}}
_C_initial = {"default": Fmi2Initial.calculated, "possible": {
    Fmi2Initial.exact, Fmi2Initial.approx, Fmi2Initial.calculated}}


_D_initial = {"default": None, "possible": {None}}
_E_initial = {"default": None, "possible": {None}}

_vc_combinations[Fmi2Variability.constant][Fmi2Causality.parameter] = {
    "err": _a_error}
_vc_combinations[Fmi2Variability.constant][Fmi2Causality.calculatedParameter] = {
    "err": _a_error}
_vc_combinations[Fmi2Variability.constant][Fmi2Causality.input] = {
    "err": _a_error}
_vc_combinations[Fmi2Variability.constant][Fmi2Causality.output] = {
    "err": None, "initial": _A_initial}
_vc_combinations[Fmi2Variability.constant][Fmi2Causality.local] = {
    "err": None, "initial": _A_initial}
_vc_combinations[Fmi2Variability.constant][Fmi2Causality.independent] = {
    "err": _c_error}


_vc_combinations[Fmi2Variability.fixed][Fmi2Causality.parameter] = {
    "err": None, "initial": _A_initial}
_vc_combinations[Fmi2Variability.fixed][Fmi2Causality.calculatedParameter] = {
    "err": None, "initial": _B_initial}
_vc_combinations[Fmi2Variability.fixed][Fmi2Causality.input] = {
    "err": _d_error}
_vc_combinations[Fmi2Variability.fixed][Fmi2Causality.output] = {
    "err": _e_error}
_vc_combinations[Fmi2Variability.fixed][Fmi2Causality.local] = {
    "err": None, "initial": _B_initial}
_vc_combinations[Fmi2Variability.fixed][Fmi2Causality.independent] = {
    "err": _c_error}

_vc_combinations[Fmi2Variability.tunable][Fmi2Causality.parameter] = {
    "err": None, "initial": _A_initial}
_vc_combinations[Fmi2Variability.tunable][Fmi2Causality.calculatedParameter] = {
    "err": None, "initial": _B_initial}
_vc_combinations[Fmi2Variability.tunable][Fmi2Causality.input] = {
    "err": _d_error}
_vc_combinations[Fmi2Variability.tunable][Fmi2Causality.output] = {
    "err": _e_error}
_vc_combinations[Fmi2Variability.tunable][Fmi2Causality.local] = {
    "err": None, "initial": _B_initial}
_vc_combinations[Fmi2Variability.tunable][Fmi2Causality.independent] = {
    "err": _c_error}


_vc_combinations[Fmi2Variability.discrete][Fmi2Causality.parameter] = {
    "err": _b_error}
_vc_combinations[Fmi2Variability.discrete][Fmi2Causality.calculatedParameter] = {
    "err": _b_error}
_vc_combinations[Fmi2Variability.discrete][Fmi2Causality.input] = {
    "err": None, "initial": _D_initial}
_vc_combinations[Fmi2Variability.discrete][Fmi2Causality.output] = {
    "err": None, "initial": _C_initial}
_vc_combinations[Fmi2Variability.discrete][Fmi2Causality.local] = {
    "err": None, "initial": _C_initial}
_vc_combinations[Fmi2Variability.discrete][Fmi2Causality.independent] = {
    "err": _c_error}

_vc_combinations[Fmi2Variability.continuous][Fmi2Causality.parameter] = {
    "err": _b_error}
_vc_combinations[Fmi2Variability.continuous][Fmi2Causality.calculatedParameter] = {
    "err": _b_error}
_vc_combinations[Fmi2Variability.continuous][Fmi2Causality.input] = {
    "err": None, "initial": _D_initial}
_vc_combinations[Fmi2Variability.continuous][Fmi2Causality.output] = {
    "err": None, "initial": _C_initial}
_vc_combinations[Fmi2Variability.continuous][Fmi2Causality.local] = {
    "err": None, "initial": _C_initial}
_vc_combinations[Fmi2Variability.continuous][Fmi2Causality.independent] = {
    "err": None, "initial": _D_initial}


def get_default_initial(variability: Fmi2Variability, causality: Fmi2Causality):

    if(validate_vc(variability, causality) is not None):
        raise Exception(
            f"Combinations of variability: {variability} and causality: {causality} is not allowed!")

    return _vc_combinations[variability][causality]["initial"]["default"]


def get_possible_initial(variability: Fmi2Variability, causality: Fmi2Causality):
    """ Returns the set of initial types that are valid for the combination of specific variability and causality.
    """
    if(validate_vc(variability, causality) is not None):
        raise Exception(
            f"Combinations of variability: {variability} and causality: {causality} is not allowed!")

    return _vc_combinations[variability][causality]["initial"]["possible"]






def validate_vc(variability: Fmi2Variability, causality: Fmi2Causality):
    """Validate combinations of variablity and causality

    Arguments:
        variability {Fmi2Variability} -- [description]
        causality {Fmi2Causality} -- [description]

    Returns:
        [type] -- [description]
    """
    return _vc_combinations[variability][causality]["err"]



import argparse
import sys
from os.path import join, dirname, realpath, normpath

from pyfmu.builder.generate import create_project
from pyfmu.builder.export import export_project
from pyfmu.builder.validate import validate


def config_generate_subprogram(subparsers: argparse.ArgumentParser) -> None:
    parser_gen = subparsers.add_parser(
        'generate', help="Generate new Python projects from scratch or based on existing resources such as Model Descriptions and reference FMUs.")

    file_project_group = parser_gen.add_mutually_exclusive_group(required=True)

    file_project_group.add_argument(
        '--path', '-p', type=str, help="Path to the project that will be generated")

    file_project_group.add_argument(
        '--file', '-f', type=str, help="Path to either a model description file or a FMU archive")

    parser_gen.add_argument(
        '--name', '-n', type=str, required=False, help="Name of the project to be generated. If not specified this information will be extracted from either the path or a model description")


def config_export_subprogram(subparsers: argparse.ArgumentParser) -> None:
    parser_export = subparsers.add_parser(
        'export',
        help="Export Python projects as FMUs",
    )

    parser_export.add_argument(
        "--project", "-p", required=True, help="path to Python project")

    parser_export.add_argument(
        "--output", '-o', required=True, help="output path of the exported archive")

    parser_export.add_argument(
        '--overwrite', '-ow', action='store_true', help='allow overwriting of existing files')

    group_bundle = parser_export.add_argument_group('bundle')

    help_bundle_interpreter = """Bundle a Python Interpreter in the FMU. 
    This will allow standalone execution on platforms where a suitable Python interpreter is not available, at the cost of an increase in archive size.
    Note that a interpreter is bundled for each supported platform.
    """
    group_bundle.add_argument(
        "-bi", type=bool, default=False, help=help_bundle_interpreter)

    help_bundle_libs = """Bundle current Python environment Libraries in the FMU. 
    """
    group_bundle.add_argument(
        "-bl", type=bool, default=True, help=help_bundle_libs)

    help_prune_libs = """ Decrease size of the included libraries by only including those used by the FMU.
    Note that this relies on analyzing static analysis of the Python files and may fail to discover all dependencies.
    """
    group_bundle.add_argument(
        "-bp", type=bool, default=False, help=help_prune_libs)


def config_validate_subprogram(subparsers: argparse.ArgumentParser) -> None:

    parser_validate = subparsers.add_parser(
        'validate', help="Static and functional verification of fmu archives")
    parser_validate.add_argument(
        'fmu', help='Path to the FMU. This may either be an zip archive or an uncompressed version of the archive')
    parser_validate.add_argument(
        '--fmpy', action='store_true', help='validate the fmu using fmpy')
    parser_validate.add_argument(
        '--vdmcheck', action='store_true', help='validate the fmu using vdmcheck')


def handle_generate(args):

    from os.path import join, curdir, dirname, basename, normpath

    if(args.file is not None):
        raise Exception("Currently generation from file is not supported")

    project_path = join(curdir, args.path)

    main_class_name = args.name if args.name is not None else basename(
        normpath(project_path))

    create_project(project_path, main_class_name)


def handle_export(args):

    project_path = args.project

    archive_path = args.output

    export_project(project_path, archive_path)


def handle_validate(args):

    fmu = args.fmu
    validate(fmu)


def main():

    description_text = """Utility program to facility the development of Functional Mock-up Units (FMUs) using Python code.
    """

    example_text = '''examples:

these are common commands for various situations:

generating a new Python project from scratch
    python py2fmu generate -n engine

generating a new Python project based on a model description file or reference FMU.
    python py2fmu generate -f modelDescription.xml
    python py2fmu generate -f engine.fmu

exporting a Python project as an FMU
    python py2fmu export -p engine
    python py2fmu export -p engine -c MyEngineClass

validating an fmu
    python py2fmu validate engine.fmu --fmpy --vdmcheck --fmicheck

re-configuring an existing project
    python py2fmu configure -p engine --bundle-interpreter
    '''

    parser = argparse.ArgumentParser(
        epilog=example_text, formatter_class=argparse.RawDescriptionHelpFormatter, description=description_text)

    subparsers = parser.add_subparsers(dest="subprogram", required=True)

    config_generate_subprogram(subparsers)
    config_export_subprogram(subparsers)
    config_validate_subprogram(subparsers)

    args = parser.parse_args()

    if(args.subprogram == "generate"):
        handle_generate(args)
    elif(args.subprogram == "export"):
        handle_export(args)
    elif(args.subprogram == 'validate'):
        handle_validate(args)
    else:
        raise Exception("Not implemented")



import os
from os.path import dirname, join
from tempfile import mkdtemp
from shutil import rmtree, copytree
from pathlib import Path
import stat


from pyfmu.builder.export import export_project, PyfmuProject, PyfmuArchive



_correct_example = {
    'Adder',
    'ConstantSignalGenerator',
    'SineGenerator',
    'LoggerFMU',
    "BicycleKinematic",
    "LivePlotting"
}

_incorrect_examples = {
    'NoneReturner'
}


def get_example_directory() -> Path:
    """
    Returns the path to the example projects
    """
    p = Path(__file__).parent.parent.parent.parent / 'examples' / 'projects'
    if(not p.is_dir()):
        raise FileNotFoundError('Expected example directory : {p} does not appear to exist. Ensure the directory exists at the specified path or update this function.')
    
    return p 

def get_incorrect_examples():
    """Returns the set of examples that are expected to fail

    Returns:
        [Set] -- The set of all examples which are designed to cause faults.
    """

    return _incorrect_examples

def get_correct_examples():
    """Returns the set of examples that are expected to work correctly.

    Returns:
        [Set] -- The set of all examples which should function correctly.
    """
    return _correct_example

def get_all_examples():
    """Returns the set of all available examples including the ones that are expected to fail.

    Returns:
        [Set] -- The set of all available examples
    """
    return get_correct_examples() | get_incorrect_examples()



def get_example_project(name: str) -> Path:
    """ Gets the path to a specific example project, to get an fmu use get_example_fmu(...)
    """
    if(name not in get_all_examples()):
        raise ValueError(
            f'Failed to resolve the path to the example project. The project {name} does not exist.')

    p = get_example_directory() / name

    return p


def get_exported_example_project(name: str):
    """ Gets the path to a specific example project, to get an fmu use get_example_fmu(...)
    """
    projects_dir = join(dirname(__file__), "export")
    return join(projects_dir, name)


class ExampleProject():
    """Wrapper that encapsulates the creation of example projects used for automatic testing.

    Wrapper that encapsulates the exporting of example archives used for automatic testing.

    This allows them to be accessed using a with statements.
    Attributes such as the path to the object can be accessed through the archive object which is returned.
    ```
    with ExampleProject('Adder') as p:
        print(p.modelDescription)
        ...
    ```

    The example project generated as a copy of an example project inside a temporary folder, which is freed after the with statement terminates.   
    Modifications of properties such as its model description can be made through its properties:
    ```
    with ExampleProject('Adder') as p:
        p.modelDescription = "invalid model description"
        ...
    ```
    """

    def __init__(self, project_name: str):

        if(project_name not in get_all_examples()):
            raise ValueError(
                f'Unable to read the example project. The specified project {project_name} could not be found.')

        project_path = get_example_project(project_name)

        # copy project to temporary directory
        self.tmpdir = mkdtemp()
        outdir = (Path(self.tmpdir) / project_name)
        copytree(project_path, outdir)

        # instantiate object representation of project
        self.project = PyfmuProject.from_existing(outdir)

    def __enter__(self) -> PyfmuProject:
        return self.project

    def __exit__(self, exception_type, exception_value, traceback):
        rmtree(self.tmpdir)


class ExampleArchive():
    """Wrapper that encapsulates the exporting of example archives used for automatic testing.

    This allows them to be accessed using a with statements.
    Attributes such as the path to the object can be accessed through the archive object which is returned.
    ```
    with ExampleArchive('Adder') as p:
        print(p.modelDescription)
        ...
    ```

    The example project is exported as an FMU to a temporary folder, which is automatically freed after the with statement terminates.   
    """

    def __init__(self, project_name: str):

        self.tmpdir = mkdtemp()

        outdir = Path(self.tmpdir) / project_name

        with ExampleProject(project_name) as p:
            self.archive = export_project(p, outdir, store_compressed=False)

    def __enter__(self) -> PyfmuArchive:
        return self.archive

    def __exit__(self, exception_type, exception_value, traceback):

        import platform

        # TODO
        # Currently the clean up mechanism does not work with FMPy on windows
        # It seems like the dll is still loaded when the function is called, resulting in a windows "access denied" error when trying to delete it.
        if(platform.system() == "Windows"):
            return

        for root, dirs, files in os.walk(self.tmpdir, topdown=False):
            for name in files:
                filename = os.path.join(root, name)
                os.chmod(filename, stat.S_IWUSR)
                os.remove(filename)
            for name in dirs:
                os.rmdir(os.path.join(root, name))



"""Defines logging related functionality
"""
from enum import Enum
from typing import Iterable, List

from pyfmu.fmi2 import Fmi2Status

# log category which the pyfmu framework logs to
_internal_log_catergory = 'pyfmu'
# default parameter for events
_default_category = 'events'


class Fmi2StdLogCats(Enum):
    """Standard log categories defined in the FMI2 specification.
    """
    logEvents = "logEvents"
    logSingularLinearSystems = "logSingularLinearSystems"
    logNonlinearSystems = "logNonlinearSystems"
    logDynamicStateSelection = "logDynamicStateSelection"
    logStatusWarning = "logStatusWarning"
    logStatusDiscard = "logStatusDiscard"
    logStatusError = "logStatusError"
    logStatusFatal = "logStatusFatal"
    logStatusPending = "logStatusPending"
    logAll = "logAll"


class Fmi2LogMessage:
    """Represents an log message as defined by FMI2
    """

    def __init__(self, status: Fmi2Status, category: str, message: str):
        """Creates a new log message

        Arguments:
            status {Fmi2Status} -- status of the FMU when the logging was performed.
            category {str} -- category of the log message.
            message {str} -- the message describing the logged event.

        Examples:

            ```
            Fmi2LogMessage(Fmi2Status.OK,"misc","foo() was called")
            ```
        """
        if(not isinstance(status, Fmi2Status)):
            raise RuntimeError("")

        if(not isinstance(category, str)):
            raise RuntimeError("")

        if(not isinstance(message, str)):
            raise RuntimeError()

        self.status: Fmi2Status = status
        self.category: str = category
        self.message: str = message


class Fmi2Logger():

    def __init__(self, callback=None):

        self._callback = callback
        self._log_stack: List[Fmi2LogMessage] = []
        self._categories_to_predicates = {}
        self._active_categories = set()

    def set_active_log_categories(self, logging_on: bool, categories: Iterable[str]):

        try:
            categories = set(categories)

        except Exception as e:
            msg = f"""Failed setting the debug categories.
            The argument categories could not be converted to a set, value was {categories}.
            Exception thrown by python was:
            {e}
            """
            self.log(Fmi2Status.warning, _internal_log_catergory, msg)
            return Fmi2Status.warning

        
        if(len(categories) == 0):
            msg = "Failed setting debug categories, list of categories is empty."
            self.log(Fmi2Status.warning, _internal_log_catergory, msg)

        not_strings = [c for c in categories if not isinstance(c, str)]
        strings = categories.difference(not_strings)

        is_all_strings = (len(not_strings) == 0)

        if(not is_all_strings):
            msg = f"""Not all elements of argument categories are strings.
            Non-string values are: {is_all_strings}
            string values are: {strings}
            Continuing using only valid categories.
            """

        if(logging_on):
            self._active_categories = self._active_categories.union(categories)
        else:
            self._active_categories = self._active_categories.difference(
                categories)


    def register_log_category(self, category: str, aliases=None, predicate=None) -> None:
        """Registers a new log category

        Arguments:
            category {str} -- [description]

        Keyword Arguments:
            aliases {[type]} -- [description] (default: {None})
            predicate {[type]} -- [description] (default: {None})
        """

        if(aliases != None and predicate != None):
            raise ValueError(
                'Unable to register log category, both a set of aliases and predicate were supplied.')

        # if neither aliases or predicate is registed log category and category are assocaited
        if(aliases == None and predicate == None):
            def default_predicated(msg: Fmi2LogMessage):
                nonlocal category
                return msg.category == category

            predicate = default_predicated

        if(aliases):

            def alias_predicate(msg: Fmi2LogMessage):
                nonlocal aliases
                return msg.category in aliases

            predicate = alias_predicate

        self._categories_to_predicates[category] = predicate

    def log(self, message: str,  category: str = _default_category, status=Fmi2Status.ok) -> None:


        if(category == None):
            category = _default_category
            
        msg = Fmi2LogMessage(status, category, message)
        

        # log only for the active categories
        activate_categories_to_predicates = { c:p for c,p in self._categories_to_predicates.items() if c in self._active_categories}

        for p in activate_categories_to_predicates.values():
            
            if(p(msg) == True):
                self._log_stack.append(msg)

                # testing callback
                if(self._callback):
                    self._callback(msg)

                break

    def register_standard_categories(self, categories: Iterable[str]) -> None:

        try:
            categories = set(categories)
        except Exception as e:
            raise ValueError(
                f'Unable to register standard log categories. The specified categories could not be converted to a set.') from e

        def events_predicate(msg: Fmi2LogMessage):
            c = msg.category.lower()
            matches = {'event', 'events'}
            return c in matches

        def sls_predicate(msg: Fmi2LogMessage):
            c = msg.category.lower()
            matches = {'singularlinearsystem', 'singularlinearsystems', 'sls'}
            return c in matches

        def nls_predicate(msg: Fmi2LogMessage):
            c = msg.category.lower()
            matches = {'nonlinearsystem', 'nonlinearsystems', 'nls'}
            return c in matches

        def dss_predicate(msg: Fmi2LogMessage):
            c = msg.category.lower()
            matches = {'dynamicstateselection', 'dss'}
            return c in matches

        def warning_predicate(msg: Fmi2LogMessage):
            return msg.status == Fmi2Status.warning

        def discard_predicate(msg: Fmi2LogMessage):
            return msg.status == Fmi2Status.discard

        def error_predicate(msg: Fmi2LogMessage):
            return msg.status == Fmi2Status.error

        def fatal_predicate(msg: Fmi2LogMessage):
            return msg.status == Fmi2Status.fatal

        def pending_predicate(msg: Fmi2LogMessage):
            return msg.status == Fmi2Status.pending

        def all_predicate(_: Fmi2LogMessage):
            return True

        predicates = {
            "logEvents"                 : events_predicate,
            "logSingularLinearSystems"  : sls_predicate,
            "logNonlinearSystems"       : nls_predicate,
            "logDynamicStateSelection"  : dss_predicate,
            "logStatusWarning"          : warning_predicate,
            "logStatusDiscard"          : discard_predicate,
            "logStatusError"            : error_predicate,
            "logStatusFatal"            : fatal_predicate,
            "logStatusPending"          : pending_predicate,
            "logAll"                    : all_predicate
        }

        predicate_matches = {cat: pred for cat,
                             pred in predicates.items() if cat in categories}

        self._categories_to_predicates = {
            **self._categories_to_predicates, **predicate_matches}

    def register_all_standard_categories(self) -> None:
        """Convenience method used to register all standard FMI2 log categories
        """
        self.register_standard_categories([
            "logEvents",
            "logSingularLinearSystems",
            "logNonlinearSystems",
            "logDynamicStateSelection",
            "logStatusWarning",
            "logStatusWarning",
            "logStatusError",
            "logStatusFatal",
            "logStatusPending",
            "logAll",
        ])

    def pop_messages(self, n: int) -> List[Fmi2LogMessage]:
        """Pops the top n messages of the message stack

        Arguments:
            n {int} -- number of messages to pop

        Returns:
            List[Fmi2LogMessage] -- List of messages
        """

        messages = self._log_stack[-n:]
        self._log_stack = self._log_stack[:-n]

        return messages

    def __len__(self):
        return len(self._log_stack)



from pyfmu.fmi2 import Fmi2Slave, Fmi2DataTypes, Fmi2Causality, Fmi2Variability
from pyfmu.builder import extract_model_description_v2

class Adder(Fmi2Slave):
    
    def __init__(self):
        super().__init__("Adder")

        self.register_variable("a", data_type=Fmi2DataTypes.real, causality = Fmi2Causality.input, start=0)
        self.register_variable("b", data_type=Fmi2DataTypes.real, causality = Fmi2Causality.input, start=0)
        self.register_variable("c", data_type=Fmi2DataTypes.real,causality = Fmi2Causality.output)

        

class SineGenerator(Fmi2Slave):

    def __init__(self):
        super().__init__('SineGenerator')
        self.register_variable('amplitude',data_type = Fmi2DataTypes.real, causality= Fmi2Causality.parameter, start=1)
        self.register_variable('frequency', data_type = Fmi2DataTypes.real, causality=Fmi2Causality.parameter, start=1)
        self.register_variable('phase', data_type = Fmi2DataTypes.real, causality=Fmi2Causality.parameter, start=0)
        self.register_variable('y', data_type = Fmi2DataTypes.real, causality=Fmi2Causality.output)


"""This file contains tests related to how the Python FMU manages resources.

For ease of testing fmpy is used as a cosimulation engine.
"""
from ctypes import cdll
import platform

import pytest
from fmpy.simulation import simulate_fmu, FMU2Slave
from fmpy.model_description import read_model_description

from pyfmu.tests import get_all_examples, get_correct_examples, get_incorrect_examples, ExampleArchive
from pyfmu.resources import Resources



def test_shared_library_can_be_loaded():

    sys = platform.system()

    for pname in get_all_examples():

        with ExampleArchive(pname) as archive:

            if(sys == 'Windows'):
                p = archive.wrapper_win64
            elif(sys == 'Linux'):
                p = archive.wrapper_linux64
            else:
                raise NotImplementedError(f'Not implemented for platform {sys}')

            p = str(p.resolve())

            cdll.LoadLibrary(p)


def test_singleInstantiation_canSimulate():

    for pname in get_correct_examples():

        with ExampleArchive(pname) as archive:

            # seems like fmpy does not accept Path objects
            path = str(archive.root.resolve())

            simulate_fmu(path)


def test_multipleExports_canSimulate():

    for pname in get_correct_examples():

        with ExampleArchive(pname) as archive:

            print(archive.model_description)
            # seems like fmpy does not accept Path objects
            path = str(archive.root.resolve())

            simulate_fmu(path)



def test_multipleInstantiationsAllDifferentInstanceNames_canSimulate():

    for idx, pname in enumerate(get_correct_examples()):

        with ExampleArchive(pname) as archive:

            print(archive.model_description)
            # seems like fmpy does not accept Path objects
            path = str(archive.root)

            md = read_model_description(str(archive.root))

            instance_a = str(idx) + 'a'
            instance_b = str(idx) + 'b'

            fmu_a = FMU2Slave(guid=md.guid,
                              unzipDirectory=archive.root,
                              modelIdentifier=md.coSimulation.modelIdentifier,
                              instanceName=instance_a,
                              fmiCallLogger=None)

            fmu_b = FMU2Slave(guid=md.guid,
                              unzipDirectory=archive.root,
                              modelIdentifier=md.coSimulation.modelIdentifier,
                              instanceName=instance_b,
                              fmiCallLogger=None)

            fmu_a.instantiate()
            fmu_b.instantiate()


def atest_identicalNamesSameTypes_throws():

    with pytest.raises(Exception):
        with ExampleArchive('Adder') as a, ExampleArchive('Adder') as b:

            md_a = read_model_description(str(b.root))
            md_b = read_model_description(str(a.root))

            instance_a = 'a'
            instance_b = 'a'

            fmu_a = FMU2Slave(guid=md_a.guid,
                              unzipDirectory=a.root,
                              modelIdentifier=md_a.coSimulation.modelIdentifier,
                              instanceName=instance_a,
                              fmiCallLogger=None)

            fmu_b = FMU2Slave(guid=md_b.guid,
                              unzipDirectory=b.root,
                              modelIdentifier=md_b.coSimulation.modelIdentifier,
                              instanceName=instance_b,
                              fmiCallLogger=None)

            fmu_a.instantiate()
            fmu_b.instantiate()


def atest_identicalNamesDifferentTypes_throws():

    with pytest.raises(Exception):
        with ExampleArchive('Adder') as a, ExampleArchive('ConstantSignalGenerator') as b:

            md_a = read_model_description(str(b.root))
            md_b = read_model_description(str(a.root))

            instance_a = 'a'
            instance_b = 'a'

            fmu_a = FMU2Slave(guid=md_a.guid,
                              unzipDirectory=a.root,
                              modelIdentifier=md_a.coSimulation.modelIdentifier,
                              instanceName=instance_a,
                              fmiCallLogger=None)

            fmu_b = FMU2Slave(guid=md_b.guid,
                              unzipDirectory=b.root,
                              modelIdentifier=md_b.coSimulation.modelIdentifier,
                              instanceName=instance_b,
                              fmiCallLogger=None)

            fmu_a.instantiate()
            fmu_b.instantiate()


def test_Adder():
    with ExampleArchive('Adder') as a:

        md_a = read_model_description(str(a.root))

        instance_a = 'a'

        fmu_a = FMU2Slave(guid=md_a.guid,
                            unzipDirectory=a.root,
                            modelIdentifier=md_a.coSimulation.modelIdentifier,
                            instanceName=instance_a,
                            fmiCallLogger=None)


        fmu_a.instantiate()
        # set input a
        fmu_a.setReal([1],[1])
        # set input b
        fmu_a.setReal([2],[2])

        fmu_a.doStep(0,1)

        # read output
        s = fmu_a.getReal([0])[0]
        assert s == 3

def test_bicycle():
    with ExampleArchive('BicycleKinematic') as a:

        md_a = read_model_description(str(a.root))


        fmu_a = FMU2Slave(guid=md_a.guid,
                            unzipDirectory=a.root,
                            modelIdentifier=md_a.coSimulation.modelIdentifier,
                            instanceName='a',
                            fmiCallLogger=None)

        fmu_a.instantiate()
        fmu_a.doStep(0,1)




from os.path import join, dirname

from pyfmu.tests import get_example_project, get_all_examples, ExampleArchive
from pyfmu.builder import validate, validate_modelDescription
from fmpy import simulate_fmu


def test_validate_md_VDMCheck(tmpdir):

    for pname in get_all_examples():

        with ExampleArchive(pname) as archive:
            results = validate_modelDescription(
                archive.model_description, use_vdmcheck=True)
            assert(results.valid == True)




from enum import Enum
from typing import Union

class Fmi2Causality(Enum):
    """ Defines the causality of the variable, that is whether it is an input, output, parameter, etc.

    Values:
        * parameter: a value set by the environment which remains constant during the simulation.
        * calculatedParameter: ?
        * input: An input from another model.
        * output: An output to other models.
        * local: ?
        * independent: Independent variable, typically time. At most one variable of an FMU may be declared as independent.

    """
    parameter = "parameter"
    calculatedParameter = "calculatedParameter"
    input = "input"
    output = "output"
    local = "local"
    independent = "independent"

class Fmi2DataTypes(Enum):
    """Defines the type of a variable.

    Values:
        * real: a real numbered value.
        * integer: a interger value.
        * boolean: a boolean value.
        * string: a text string.
    """
    real = "Real"
    integer = "Integer"
    boolean = "Boolean"
    string = "String"

class Fmi2Initial(Enum):
    """Defines how the initial value of a variable is initialized.
    
    Values:
        exact: The variable is initialised with the provided start value.
        calculated: The variable is defined based on other variables during initialisation. 
        "approx": The variable is defined based on the an iteration of an algebraic loop with the provided start value.
    """
    exact = "exact"
    approx = "approx"
    calculated = "calculated"

class Fmi2Variability(Enum):
    """ Defines the time instants where the variable can change value.

    Values:
        * constant: The variable never changes.
        * fixed: The variable never changes after initialization, specifically after exit_initialization_mode has been called.
        * tunable: ?
        * discrete : ?
        * continuous : No restriction on when the variable can change. 
    """
    constant = "constant"
    fixed = "fixed"
    tunable = "tunable"
    discrete = "discrete"
    continuous = "continuous"

class Fmi2Status(Enum):
    """Represents the status of the FMU or the results of function calls.

    Values:
        * ok: all well
        * warning: an issue has arisen, but the computation can continue.
        * discard: an operation has resulted in invalid output, which must be discarded
        * error: an error has ocurred for this specific FMU instance.
        * fatal: an fatal error has ocurred which has corrupted ALL FMU instances.
        * pending: indicates that the FMu is doing work asynchronously, which can be retrived later.

    Notes:
        FMI section 2.1.3    
    
    """
    ok = 0
    warning = 1
    discard = 2
    error = 3
    fatal = 4
    pending = 5



from setuptools import setup, find_packages
setup(
    name="pyfmu",
    version="0.0.1",
    description='A set of tools for developing functional-mockup-units (FMUs) using the full capabilities of Python.',
    author='INTO-CPS Association',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    project_urls={
        'Bug Tracker': 'https://github.com/INTO-CPS-Association/pyfmu/issues',
        'Documentation': 'https://into-cps-application.readthedocs.io/en/latest/submodules/pyfmu/docs/index.html',
        'Source Code': 'https://github.com/INTO-CPS-Association/pyfmu'
    },

    install_requires=[
        'Jinja2'
    ],
    extras_require={

        "docs": ["Sphinx", "recommonmark", "sphinx_rtd_theme", "sphinx-autoapi"],
        "tests": ["numpy", "pandas", "pytest", "fmpy", "pyqtgraph", "scipy", "PyQt5"],
        "gui": []
    },

    # resources needed by the CLI to generate and export
    package_data={
        "pyfmu": [
            "resources/templates/*.j2",
            "resources/wrapper/binaries/*/*[.so|.dll]",
            "resources/validation/*"
        ]
    },
    include_package_data=True,

    python_requires='>=3',

    entry_points={'console_scripts': ['pyfmu=pyfmu.pyfmu_cli:main']}


)



from pyfmu.fmi2 import Fmi2Slave, Fmi2DataTypes, Fmi2Causality, Fmi2Variability, Fmi2Status, Fmi2Initial,Fmi2StdLogCats


class Adder(Fmi2Slave):
    
    def __init__(self):
        super().__init__("Adder")

        self.register_variable("a", data_type=Fmi2DataTypes.real, causality = Fmi2Causality.input, start=0)
        self.register_variable("b", data_type=Fmi2DataTypes.real, causality = Fmi2Causality.input, start=0)
        self.register_variable("c", data_type=Fmi2DataTypes.real,causality = Fmi2Causality.output)

class Dummy(Fmi2Slave):

    def __init__(self):
        super().__init__("Dummy")

class Tests_fmi2Slave:
    
    def test_register_variable_implicitly_defines_attribute(self):

        a = Adder()

        assert(hasattr(a,'a'))
        assert(hasattr(a,'b'))
        assert(hasattr(a,'c'))

        assert(a.a == 0)
        assert(a.b == 0)


def test_inputsUseStartValue():
    
    d = Dummy()
    
    start = 10
    vr = 0

    d.register_variable("a",data_type = Fmi2DataTypes.real,causality=Fmi2Causality.input, start=start,value_reference=vr)

    result = [0]

    d.__get_real__([vr],result)
    assert(result[0] == start)

def test_parametersUseStartValue():
    d = Dummy()

    start = 10
    vr = 0

    d.register_variable("a",data_type=Fmi2DataTypes.real,variability=Fmi2Variability.fixed,causality = Fmi2Causality.parameter,start=start,value_reference=vr)

    result = [0]

    d.__get_real__([vr],result)
    assert(result[0] == start)

def test_registerVariable_acceptsStrings():

    slave = Fmi2Slave("")

    valid_combinations = {
        ('real','input','continuous',None),
        ('real','output','continuous','exact')
    }

    for data_type,causality,variability,initial in valid_combinations:
        slave.register_variable("a",data_type=data_type,causality=causality,variability=variability,initial=initial,start=0)

def test_registerVariable_infersDataTypeFromStart():

    s = Fmi2Slave("")
    s.register_variable('a',causality='input',start=1.0)    
    matches = [v for v in s.vars if v.data_type is Fmi2DataTypes.real]

    assert(len(matches) is 1)


def test_registerVariable_defaultStartValues():

    s = Fmi2Slave("")

    # exact should apply default values
    s.register_variable('re','real',initial='exact')
    s.register_variable('ie','integer',variability='discrete',initial='exact')
    s.register_variable('be','boolean',variability='discrete',initial='exact')
    s.register_variable('se','string',variability='discrete',initial='exact')

    assert(s.re == 0.0)
    assert(s.ie == 0)
    assert(s.be == False)
    assert(s.se == "")
    
    v_re = [v for v in s.vars if v.name == 're']
    v_ie = [v for v in s.vars if v.name == 'ie']
    v_be = [v for v in s.vars if v.name == 'be']
    v_se = [v for v in s.vars if v.name == 'se']
    
    assert(len(v_re) == 1)
    assert(len(v_ie) == 1)
    assert(len(v_be) == 1)
    assert(len(v_se) == 1)

    # approx should also apply default values

    s.register_variable('ra','real',initial='approx')
    s.register_variable('ia','integer',variability='discrete',initial='approx')
    s.register_variable('ba','boolean',variability='discrete',initial='approx')
    s.register_variable('sa','string',variability='discrete',initial='approx')

    assert(s.ra == 0.0)
    assert(s.ia == 0)
    assert(s.ba == False)
    assert(s.sa == "")

    v_ra = [v for v in s.vars if v.name == 'ra']
    v_ia = [v for v in s.vars if v.name == 'ia']
    v_ba = [v for v in s.vars if v.name == 'ba']
    v_sa = [v for v in s.vars if v.name == 'sa']
    
    assert(len(v_ra) == 1)
    assert(len(v_ia) == 1)
    assert(len(v_ba) == 1)
    assert(len(v_sa) == 1)
    


    

    


# test logging functions used by the wrapper


def test_setDebugLogging():

    fmu = Fmi2Slave("a")

    fmu.log("test")
    
    assert(fmu.__get_log_size__() == 0)

    fmu.__set_debug_logging__(True,["logAll"])

    fmu.log("test")

    assert(fmu.__get_log_size__() == 1)
    

def test__get_log_size__():
    
    fmu = Fmi2Slave("logger",standard_log_categories=True)
    fmu.__set_debug_logging__(True,["logAll"])

    fmu.log("test 1")
    
    assert(fmu.__get_log_size__() == 1)
    
    fmu.log("test 2")

    assert(fmu.__get_log_size__() == 2)


def test__get_log_messages__():

    fmu = Fmi2Slave("logger", standard_log_categories=True)
    fmu.__set_debug_logging__(True,["logAll"])

    fmu.log("test",category="a",status=Fmi2Status.ok)

    ms = fmu.__pop_log_messages__(1)

    (status,category,message) = ms[0]

    assert(status == Fmi2Status.ok.value)
    assert(isinstance(status,int))
    assert(category == "a")
    assert(message == 'test')



from pathlib import Path


_resources = None

_wrapper_name = 'pyfmu'

class Resources:
    """Singleton object representing static resources.


    For example to get path to pyfmu directory:
    ```
    pyfmu_dir = Resources.get().pyfmu_dir
    ```

    Do not move the file as the resources must are located relative to this file
    """

    @staticmethod
    def get():

        global _resources

        if(_resources is None):
            _resources = Resources()

        return _resources

    def __init__(self):
        self.root = Path(__file__).parent
        self.pyfmu_dir = self.root / 'pyfmu'
        self.templates_dir = self.root / 'templates'
        self.scriptTemplate_path = self.templates_dir / 'fmu.py.j2'

        # binaries
        self.binaries_dir: Path = self.root / 'wrapper' / 'binaries'
        self.wrapper_win64 = self.binaries_dir / 'win64' / (_wrapper_name + '.dll')
        self.wrapper_linux64 = self.binaries_dir / 'linux64' / (_wrapper_name + '.so')


        self.VDMCheck2_jar = self.root / 'validation' / 'vdmcheck-0.0.2' / 'fmi2vdm-0.0.2.jar'
        self.VDMCheck3_jar = self.root / 'validation' / 'vdmcheck-0.0.3' / 'fmi2vdm-0.0.3.jar'


from pyfmu.fmi2 import Fmi2Logger, Fmi2LogMessage, Fmi2StdLogCats,Fmi2Status,Fmi2Slave

def test_categoryNotRegistered_notLogged():
    
    logged = False

    def logging_callback(msg : Fmi2LogMessage):
        nonlocal logged 
        logged = True

    logger = Fmi2Logger(callback=logging_callback)
    logger.log("test")

    assert(not logged)

def test_customCategoryRegistered_logged():

    c = "test"

    logger = Fmi2Logger()
    logger.register_log_category(c)
    logger.set_active_log_categories(True,[c])
    
    assert(len(logger) == 0)
    
    logger.log("hello world!",category=c)

    assert(len(logger) == 1)

def test_registerAlias_logged():
    
    logger = Fmi2Logger()
    logger.register_log_category('ab',aliases={'a','b'})
    logger.set_active_log_categories(True,['ab'])

    logger.log('test','a')
    logger.log('test','b')
    
    assert(len(logger) == 2)

def test_logAllPredicate_allLogged():

   
    def log_all_predicate(category : str) -> bool:
        return True

    logger = Fmi2Logger()
    logger.register_log_category('all',predicate=log_all_predicate)
    logger.set_active_log_categories(True,['all'])
    
    # different categories
    logger.log('test','a')
    logger.log('test','b')

    # different status
    logger.log('test',status=Fmi2Status.ok)
    logger.log('test',status=Fmi2Status.warning)

    assert(len(logger))

# standard categories

def test_logEvent():


    cs = ["logEvents"]

    logger = Fmi2Logger()
    logger.set_active_log_categories(True,cs)
    logger.register_standard_categories(cs)

    # every log message is treated as an event?
    logger.log('test',status=Fmi2Status.ok)
    logger.log('test',status=Fmi2Status.warning)
    logger.log('test',status=Fmi2Status.error)
    logger.log('test',status=Fmi2Status.discard)
    logger.log('test',status=Fmi2Status.fatal)
    logger.log('test',status=Fmi2Status.pending)
    
    assert(len(logger) == 6)

def test_logSingularLinearSystems():
   
    cs = ["logSingularLinearSystems"]

    logger = Fmi2Logger()
    logger.register_standard_categories(cs)
    logger.set_active_log_categories(True,cs)

    # category aliases
    logger.log('test',category="SingularLinearSystem")
    logger.log('test',category="singularlinearsystem")
    logger.log('test',category="sls")

    # ignore everything else
    logger.log('test',status=Fmi2Status.ok)
    logger.log('test',status=Fmi2Status.warning)
    logger.log('test',status=Fmi2Status.error)
    logger.log('test',status=Fmi2Status.discard)
    logger.log('test',status=Fmi2Status.fatal)
    logger.log('test',status=Fmi2Status.pending)
    
    assert(len(logger))

def test_logNonLinearSystems():
    
    cs = ["logNonlinearSystems"]

    logger = Fmi2Logger()
    logger.register_standard_categories(cs)
    logger.set_active_log_categories(True,cs)

    # category aliases
    logger.log('test',category="NonLinearSystems")
    logger.log('test',category="nonlinearsystems")
    logger.log('test',category="nls")

    # ignore everything else
    logger.log('test',status=Fmi2Status.ok)
    logger.log('test',status=Fmi2Status.warning)
    logger.log('test',status=Fmi2Status.error)
    logger.log('test',status=Fmi2Status.discard)
    logger.log('test',status=Fmi2Status.fatal)
    logger.log('test',status=Fmi2Status.pending)
    
    assert(len(logger) == 3)

def test_logDynamicStateSelection():

    cs = ["logDynamicStateSelection"]

    logger = Fmi2Logger()
    logger.register_standard_categories(cs)
    logger.set_active_log_categories(True,cs)
    
    logger.log('test',category="DynamicStateSelection")
    logger.log('test',category="dynamicstateselection")
    logger.log('test',category="dss")
    
    logger.log('test',status=Fmi2Status.ok)
    logger.log('test',status=Fmi2Status.warning)
    logger.log('test',status=Fmi2Status.error)
    logger.log('test',status=Fmi2Status.discard)
    logger.log('test',status=Fmi2Status.fatal)
    logger.log('test',status=Fmi2Status.pending)
    
    assert(len(logger) == 3)

def test_logWarning():

    cs = ["logStatusWarning"]

    logger = Fmi2Logger()
    logger.register_standard_categories(cs)
    logger.set_active_log_categories(True,cs)

    logger.log('test',status=Fmi2Status.ok)
    assert(len(logger) == 0)
    logger.log('test',status=Fmi2Status.warning)
    assert(len(logger) == 1)
    logger.log('test',status=Fmi2Status.error)
    logger.log('test',status=Fmi2Status.discard)
    logger.log('test',status=Fmi2Status.fatal)
    logger.log('test',status=Fmi2Status.pending)
    assert(len(logger) == 1)

def test_logError():

    cs = ["logStatusError"]

    logger = Fmi2Logger()
    logger.register_standard_categories(cs)
    logger.set_active_log_categories(True,cs)

    logger.log('test',status=Fmi2Status.ok)
    logger.log('test',status=Fmi2Status.warning)
    assert(len(logger) == 0)
    logger.log('test',status=Fmi2Status.error)
    assert(len(logger) == 1)
    logger.log('test',status=Fmi2Status.discard)
    logger.log('test',status=Fmi2Status.fatal)
    logger.log('test',status=Fmi2Status.pending)
    assert(len(logger) == 1)

def test_logDiscard():

    cs = ["logStatusDiscard"]

    logger = Fmi2Logger()
    logger.register_standard_categories(cs)
    logger.set_active_log_categories(True,cs)

    logger.log('test',status=Fmi2Status.ok)
    logger.log('test',status=Fmi2Status.warning)
    logger.log('test',status=Fmi2Status.error)
    assert(len(logger) == 0)
    logger.log('test',status=Fmi2Status.discard)
    assert(len(logger) == 1)
    logger.log('test',status=Fmi2Status.fatal)
    logger.log('test',status=Fmi2Status.pending)
    assert(len(logger) == 1)

def test_logFatal():

    cs = ["logStatusFatal"]

    logger = Fmi2Logger()
    logger.register_standard_categories(cs)
    logger.set_active_log_categories(True,cs)

    logger.log('test',status=Fmi2Status.ok)
    logger.log('test',status=Fmi2Status.warning)
    logger.log('test',status=Fmi2Status.error)
    logger.log('test',status=Fmi2Status.discard)
    assert(len(logger) == 0)
    logger.log('test',status=Fmi2Status.fatal)
    assert(len(logger) == 1)
    logger.log('test',status=Fmi2Status.pending)
    assert(len(logger) == 1)

def test_logPending():

    cs = ["logStatusPending"]

    logger = Fmi2Logger()
    logger.register_standard_categories(cs)
    logger.set_active_log_categories(True,cs)

    logger.log('test',status=Fmi2Status.ok)
    logger.log('test',status=Fmi2Status.warning)
    logger.log('test',status=Fmi2Status.error)
    logger.log('test',status=Fmi2Status.discard)
    logger.log('test',status=Fmi2Status.fatal)
    assert(len(logger) == 0)
    logger.log('test',status=Fmi2Status.pending)
    assert(len(logger) == 1)

def test_logAll():

    cs = ["logAll"]

    logger = Fmi2Logger()
    logger.set_active_log_categories(True,cs)
    logger.register_standard_categories(cs)
    
    
    logger.log('test',status=Fmi2Status.ok)
    logger.log('test',status=Fmi2Status.warning)
    logger.log('test',status=Fmi2Status.error)
    logger.log('test',status=Fmi2Status.discard)
    logger.log('test',status=Fmi2Status.fatal)
    logger.log('test',status=Fmi2Status.pending)
    
    assert(len(logger) == 6)

# defaults

def test_logDefaultCategory_loggedWithEvent():

    cs = ["logEvents"]

    logger = Fmi2Logger()
    logger.register_standard_categories(cs)
    logger.set_active_log_categories(True,cs)

    assert(len(logger) == 0)
    logger.log('test')
    assert(len(logger) == 1)


# message stack

def test_len_logIncreasesLength():
    
    cs = ["logAll"]

    logger = Fmi2Logger()
    logger.register_standard_categories(cs)
    logger.set_active_log_categories(True,cs)

    assert(len(logger) == 0)

    logger.log('test')

    assert(len(logger) == 1)

    logger.log('test')

    assert(len(logger) == 2)

def test_len_popDecreasesLength():
    cs = ["logAll"]
    
    logger = Fmi2Logger()
    logger.register_standard_categories(cs)
    logger.set_active_log_categories(True,cs)

    assert(len(logger) == 0)

    logger.log('test')

    assert(len(logger) == 1)

    logger.log('test')

    assert(len(logger) == 2)

    logger.pop_messages(1)

    assert(len(logger) == 1)

    logger.pop_messages(1)

    assert(len(logger) == 0)

def test_pop_returnsMessagesFifo():

    cs = ["logAll"]

    logger = Fmi2Logger()
    logger.register_standard_categories(cs)
    logger.set_active_log_categories(True,cs)

    logger.log("a")
    logger.log("b")

    messages = logger.pop_messages(2)
    messages = [m.message for m in messages]
    
    assert(messages == ['a','b'])










import pytest
from pyfmu.fmi2 import Fmi2ScalarVariable, Fmi2Initial, Fmi2DataTypes, Fmi2Causality, Fmi2Variability


def test_defaultCausality_isLocal():
    v = Fmi2ScalarVariable("a", data_type=Fmi2DataTypes.real)
    assert(v.causality == Fmi2Causality.local)


def test_defaultVariability_isContinous():
    v = Fmi2ScalarVariable("a", data_type=Fmi2DataTypes.real)
    assert(v.variability == Fmi2Variability.continuous)


def test_input_initialNotAllowed():

    with pytest.raises(Exception):
        v = Fmi2ScalarVariable("a", data_type=Fmi2DataTypes.real,
                           causality=Fmi2Causality.input, initial=Fmi2Initial.exact)

    with pytest.raises(Exception):
        v = Fmi2ScalarVariable("a", data_type=Fmi2DataTypes.real,
                           causality=Fmi2Causality.input, initial=Fmi2Initial.approx)

    with pytest.raises(Exception):
        v = Fmi2ScalarVariable("a", data_type=Fmi2DataTypes.real,
                           causality=Fmi2Causality.input, initial=Fmi2Initial.calculated)


def test_approx_startDefined_NOK():
    with pytest.raises(Exception):
        Fmi2ScalarVariable("a", data_type=Fmi2DataTypes.real,
                       initial=Fmi2Initial.approx)


def test_approx_startNotDefined_OK():
    Fmi2ScalarVariable("a", data_type=Fmi2DataTypes.real,
                   initial=Fmi2Initial.approx, start=0)


def test_initialIsCalculated_startDefined_NOK():
    with pytest.raises(Exception):
        Fmi2ScalarVariable("a", data_type=Fmi2DataTypes.real,
                       initial=Fmi2Initial.calculated, start=0)


def test_initialIsCalculated_startNotDefined_OK():
    Fmi2ScalarVariable("a", data_type=Fmi2DataTypes.real,
                   initial=Fmi2Initial.calculated)


def test_initialIsExact_startNotDefined_NOK():
    with pytest.raises(Exception):
        Fmi2ScalarVariable("a", data_type=Fmi2DataTypes.real,
                       initial=Fmi2Initial.exact)


def test_initialIsExact_startDefined_NOK():

    Fmi2ScalarVariable("a", data_type=Fmi2DataTypes.real,
                   initial=Fmi2Initial.exact, start=0)


def test_input_startNotDefined_NOK():
    with pytest.raises(Exception):
        Fmi2ScalarVariable("a", causality=Fmi2Causality.input,
                       data_type=Fmi2DataTypes.real)


def test_input_startDefined_OK():
    Fmi2ScalarVariable("a", causality=Fmi2Causality.input,
                   data_type=Fmi2DataTypes.real, start=0)


def test_dataTypeAndStartTypeMismatch_NOK():


    type_to_invalidStartValues = {
        Fmi2DataTypes.real: {
            '', True, False
        },
        Fmi2DataTypes.boolean: {
            '', 1, 0, 0.5
        },
        Fmi2DataTypes.integer: {
            '', True, False, 0.5, 1.5
        }}

  

    with pytest.raises(TypeError):
        for t, ss in type_to_invalidStartValues.items():
            for s in ss:
                Fmi2ScalarVariable("",data_type = t,causality=Fmi2Causality.input,start=s) 

# Output valid start values

def test_output_exact_startDefined_OK():
    Fmi2ScalarVariable('a', causality=Fmi2Causality.output,
                   initial=Fmi2Initial.exact, data_type=Fmi2DataTypes.real, start=0)


def test_output_exact_startUndefined_NOK():
    with pytest.raises(Exception):
        Fmi2ScalarVariable('a', causality=Fmi2Causality.output,
                       initial=Fmi2Initial.exact, data_type=Fmi2DataTypes.real)


def test_output_calculated_startDefined_NOK():
    with pytest.raises(Exception):
        Fmi2ScalarVariable('a', causality=Fmi2Causality.output,
                       initial=Fmi2Initial.calculated, data_type=Fmi2DataTypes.real, start=0)


def test_output_calculated_startUndefined_OK():
    Fmi2ScalarVariable('a', causality=Fmi2Causality.output,
                   initial=Fmi2Initial.calculated, data_type=Fmi2DataTypes.real)


def test_output_approx_startDefined_OK():
    Fmi2ScalarVariable('a', causality=Fmi2Causality.output,
                   initial=Fmi2Initial.approx, data_type=Fmi2DataTypes.real, start=0)


def test_output_approx_startUndefined_NOK():
    with pytest.raises(Exception):
        Fmi2ScalarVariable('a', causality=Fmi2Causality.output,
                       initial=Fmi2Initial.approx, data_type=Fmi2DataTypes.real)
